"""
Base abstract classes:
- BaseRecommender - the simplest base class
- Recommender - base class for models that fit on interactions
- HybridRecommender - base class for models that accept query or item features
- QueryRecommender - base class that accepts only query features, but not item features
- NeighbourRec - base class that requires interactions at prediction time
- ItemVectorModel - class for models which provides items' vectors.
    Implements similar items search.
- NonPersonalizedRecommender - base class for non-personalized recommenders
    with popularity statistics
"""

import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterable
from os.path import join
from typing import Any, Optional, Union

import numpy as np
import pandas as pd
from numpy.random import default_rng

from replay.data import Dataset, get_schema
from replay.models.common import RecommenderCommons
from replay.models.optimization import IsOptimizible
from replay.utils import PYSPARK_AVAILABLE, PandasDataFrame, SparkDataFrame
from replay.utils.session_handler import State
from replay.utils.spark_utils import SparkCollectToMasterWarning

if PYSPARK_AVAILABLE:
    from pyspark.sql import (
        Window,
        functions as sf,
    )

    from replay.utils.spark_utils import (
        convert2spark,
        cosine_similarity,
        filter_cold,
        get_top_k,
        get_top_k_recs,
        get_unique_entities,
        load_pickled_from_parquet,
        return_recs,
        save_picklable_to_parquet,
        vector_dot,
        vector_euclidean_distance_similarity,
    )


class IsSavable(ABC):
    """
    Common methods and attributes for saving and loading RePlay models
    """

    @property
    @abstractmethod
    def _init_args(self) -> dict:
        """
        Dictionary of the model attributes passed during model initialization.
        Used for model saving and loading
        """

    @property
    def _dataframes(self) -> dict:
        """
        Dictionary of the model dataframes required for inference.
        Used for model saving and loading
        """
        return {}

    @abstractmethod
    def _save_model(self, path: str) -> None:
        """
        Method for dump model attributes to disk
        """

    @abstractmethod
    def _load_model(self, path: str) -> None:
        """
        Method for loading model attributes from disk
        """


class BaseRecommender(IsSavable, IsOptimizible, RecommenderCommons, ABC):
    """Base recommender"""

    model: Any
    can_predict_cold_queries: bool = False
    can_predict_cold_items: bool = False
    fit_queries: SparkDataFrame
    fit_items: SparkDataFrame
    _num_queries: int
    _num_items: int
    _query_dim_size: int
    _item_dim_size: int

    def _fit_wrap(
        self,
        dataset: Dataset,
    ) -> None:
        """
        Wrapper for fit to allow for fewer arguments in a model.

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :return:
        """
        self.query_column = dataset.feature_schema.query_id_column
        self.item_column = dataset.feature_schema.item_id_column
        self.rating_column = dataset.feature_schema.interactions_rating_column
        self.timestamp_column = dataset.feature_schema.interactions_timestamp_column
        self.logger.debug("Starting fit %s", type(self).__name__)
        if dataset.query_features is None:
            queries = dataset.interactions.select(self.query_column).distinct()
        else:
            queries = (
                dataset.interactions.select(self.query_column)
                .union(dataset.query_features.select(self.query_column))
                .distinct()
            )
        if dataset.item_features is None:
            items = dataset.interactions.select(self.item_column).distinct()
        else:
            items = (
                dataset.interactions.select(self.item_column)
                .union(dataset.item_features.select(self.item_column))
                .distinct()
            )
        self.fit_queries = sf.broadcast(queries)
        self.fit_items = sf.broadcast(items)
        self._num_queries = self.fit_queries.count()
        self._num_items = self.fit_items.count()
        self._query_dim_size = self.fit_queries.agg({self.query_column: "max"}).first()[0] + 1
        self._item_dim_size = self.fit_items.agg({self.item_column: "max"}).first()[0] + 1
        self._fit(dataset)

    @abstractmethod
    def _fit(
        self,
        dataset: Dataset,
    ) -> None:
        """
        Inner method where model actually fits.

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :return:
        """

    def _filter_seen(
        self,
        recs: SparkDataFrame,
        interactions: SparkDataFrame,
        k: int,
        queries: SparkDataFrame,
    ):
        """
        Filter seen items (presented in interactions) out of the queries' recommendations.
        For each query return from `k` to `k + number of seen by query` recommendations.
        """
        queries_interactions = interactions.join(queries, on=self.query_column)
        self._cache_model_temp_view(queries_interactions, "filter_seen_queries_interactions")
        num_seen = queries_interactions.groupBy(self.query_column).agg(sf.count(self.item_column).alias("seen_count"))
        self._cache_model_temp_view(num_seen, "filter_seen_num_seen")

        # count maximal number of items seen by queries
        max_seen = 0
        if num_seen.count() > 0:
            max_seen = num_seen.select(sf.max("seen_count")).first()[0]

        # crop recommendations to first k + max_seen items for each query
        recs = recs.withColumn(
            "temp_rank",
            sf.row_number().over(Window.partitionBy(self.query_column).orderBy(sf.col(self.rating_column).desc())),
        ).filter(sf.col("temp_rank") <= sf.lit(max_seen + k))

        # leave k + number of items seen by query recommendations in recs
        recs = (
            recs.join(num_seen, on=self.query_column, how="left")
            .fillna(0)
            .filter(sf.col("temp_rank") <= sf.col("seen_count") + sf.lit(k))
            .drop("temp_rank", "seen_count")
        )

        # filter recommendations presented in interactions interactions
        recs = recs.join(
            queries_interactions.withColumnRenamed(self.item_column, "item")
            .withColumnRenamed(self.query_column, "query")
            .select("query", "item"),
            on=(sf.col(self.query_column) == sf.col("query")) & (sf.col(self.item_column) == sf.col("item")),
            how="anti",
        ).drop("query", "item")

        return recs

    def _filter_interactions_queries_items_dataframes(
        self,
        dataset: Optional[Dataset],
        k: int,
        queries: Optional[Union[SparkDataFrame, Iterable]] = None,
        items: Optional[Union[SparkDataFrame, Iterable]] = None,
    ):
        """
        Returns triplet of filtered `dataset`, `queries`, and `items`.
        Filters out cold entities (queries/items) from the `queries`/`items` and `dataset`
        if the model does not predict cold.
        Filters out duplicates from `queries` and `items` dataframes,
        and excludes all columns except `user_idx` and `item_idx`.

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :param k: number of recommendations for each user
        :param queries: queries to create recommendations for
            dataframe containing ``[user_idx]`` or ``array-like``;
            if ``None``, recommend to all queries from ``dataset``
        :param items: candidate items for recommendations
            dataframe containing ``[item_idx]`` or ``array-like``;
            if ``None``, take all items from ``dataset``.
            If it contains new items, ``rating`` for them will be ``0``.
        :return: triplet of filtered `dataset`, `queries`, and `items`.
        """
        self.logger.debug("Starting predict %s", type(self).__name__)
        if dataset is not None:
            query_data = queries or dataset.interactions or dataset.query_features or self.fit_queries
            interactions = dataset.interactions
        else:
            query_data = queries or self.fit_queries
            interactions = None

        queries = get_unique_entities(query_data, self.query_column)
        queries, interactions = self._filter_cold_for_predict(queries, interactions, "query")

        item_data = items or self.fit_items
        items = get_unique_entities(item_data, self.item_column)
        items, interactions = self._filter_cold_for_predict(items, interactions, "item")
        num_items = items.count()
        if num_items < k:
            message = f"k = {k} > number of items = {num_items}"
            self.logger.debug(message)

        if dataset is not None:
            dataset = Dataset(
                feature_schema=dataset.feature_schema,
                interactions=interactions,
                query_features=dataset.query_features,
                item_features=dataset.item_features,
                check_consistency=False,
            )
        return dataset, queries, items

    def _predict_wrap(
        self,
        dataset: Optional[Dataset],
        k: int,
        queries: Optional[Union[SparkDataFrame, Iterable]] = None,
        items: Optional[Union[SparkDataFrame, Iterable]] = None,
        filter_seen_items: bool = True,
        recs_file_path: Optional[str] = None,
    ) -> Optional[SparkDataFrame]:
        """
        Predict wrapper to allow for fewer parameters in models

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :param k: number of recommendations for each user
        :param queries: queries to create recommendations for
            dataframe containing ``[user_idx]`` or ``array-like``;
            if ``None``, recommend to all queries from ``interactions``
        :param items: candidate items for recommendations
            dataframe containing ``[item_idx]`` or ``array-like``;
            if ``None``, take all items from ``interactions``.
            If it contains new items, ``rating`` for them will be ``0``.
        :param user_features: user features
            ``[user_idx , timestamp]`` + feature columns
        :param item_features: item features
            ``[item_idx , timestamp]`` + feature columns
        :param filter_seen_items: flag to remove seen items from recommendations based on ``interactions``.
        :param recs_file_path: save recommendations at the given absolute path as parquet file.
            If None, cached and materialized recommendations dataframe  will be returned
        :return: cached recommendation dataframe with columns ``[user_idx, item_idx, rating]``
            or None if `file_path` is provided
        """
        dataset, queries, items = self._filter_interactions_queries_items_dataframes(dataset, k, queries, items)

        recs = self._predict(
            dataset,
            k,
            queries,
            items,
            filter_seen_items,
        )
        if filter_seen_items and dataset is not None:
            recs = self._filter_seen(recs=recs, interactions=dataset.interactions, queries=queries, k=k)

        recs = get_top_k_recs(recs, k=k, query_column=self.query_column, rating_column=self.rating_column).select(
            self.query_column, self.item_column, self.rating_column
        )

        output = return_recs(recs, recs_file_path)
        self._clear_model_temp_view("filter_seen_queries_interactions")

        self._clear_model_temp_view("filter_seen_num_seen")
        return output

    def _filter_cold_for_predict(
        self,
        main_df: SparkDataFrame,
        interactions_df: Optional[SparkDataFrame],
        entity: str,
    ):
        """
        Filter out cold entities (queries/items) from the `main_df` and `interactions_df_df`
        if the model does not predict cold.
        Warn if cold entities are present in the `main_df`.
        """
        can_predict_cold = self.can_predict_cold_queries if entity == "query" else self.can_predict_cold_items
        if can_predict_cold:
            return main_df, interactions_df

        fit_entities = self.fit_queries if entity == "query" else self.fit_items
        column = self.query_column if entity == "query" else self.item_column

        num_new, main_df = filter_cold(main_df, fit_entities, col_name=column)
        if num_new > 0:
            self.logger.info(
                "%s model can't predict cold %ss, they will be ignored",
                self,
                entity,
            )
        _, interactions_df = filter_cold(interactions_df, fit_entities, col_name=column)
        return main_df, interactions_df

    @abstractmethod
    def _predict(
        self,
        dataset: Dataset,
        k: int,
        queries: SparkDataFrame,
        items: SparkDataFrame,
        filter_seen_items: bool = True,
    ) -> SparkDataFrame:
        """
        Inner method where model actually predicts.

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :param k: number of recommendations for each user
        :param queries: queries to create recommendations for
            dataframe containing ``[user_idx]`` or ``array-like``;
            if ``None``, recommend to all queries from ``interactions``
        :param items: candidate items for recommendations
            dataframe containing ``[item_idx]`` or ``array-like``;
            if ``None``, take all items from ``interactions``.
            If it contains new items, ``rating`` for them will be ``0``.
        :param filter_seen_items: flag to remove seen items from recommendations based on ``interactions``.
        :return: recommendation dataframe
            ``[user_idx, item_idx, rating]``
        """

    def _predict_proba(
        self,
        dataset: Dataset,
        k: int,
        queries: SparkDataFrame,
        items: SparkDataFrame,
        filter_seen_items: bool = True,
    ) -> np.ndarray:
        """
        Inner method where model actually predicts probability estimates.

        Mainly used in ```OBPOfflinePolicyLearner```.

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :param k: number of recommendations for each user
        :param queries: queries to create recommendations for
            dataframe containing ``[user_idx]`` or ``array-like``;
            if ``None``, recommend to all queries from ``interactions``
        :param items: candidate items for recommendations
            dataframe containing ``[item_idx]`` or ``array-like``;
            if ``None``, take all items from ``interactions``.
            If it contains new items, ``rating`` for them will be ``0``.
        :param filter_seen_items: flag to remove seen items from recommendations based on ``interactions``.
        :return: distribution over items for each user with shape
            ``(n_users, n_items, k)``
            where we have probability for each user to choose item at fixed position(top-k).
        """

        n_users = queries.select("user_idx").count()
        n_items = items.select("item_idx").count()

        recs = self._predict(dataset, k, queries, items, filter_seen_items)

        recs = get_top_k_recs(recs, k=k, query_column=self.query_column, rating_column=self.rating_column).select(
            self.query_column, self.item_column, self.rating_column
        )

        cols = [f"k{i}" for i in range(k)]

        recs_items = (
            recs.groupBy("user_idx")
            .agg(sf.collect_list("item_idx").alias("item_idx"))
            .select([sf.col("item_idx")[i].alias(cols[i]) for i in range(k)])
        )

        action_dist = np.zeros(shape=(n_users, n_items, k))

        for i in range(k):
            action_dist[
                np.arange(n_users),
                recs_items.select(cols[i]).toPandas()[cols[i]].to_numpy(),
                np.ones(n_users, dtype=int) * i,
            ] += 1

        return action_dist

    def _get_fit_counts(self, entity: str) -> int:
        num_entities = "_num_queries" if entity == "query" else "_num_items"
        fit_entities = self.fit_queries if entity == "query" else self.fit_items
        if not hasattr(self, num_entities):
            setattr(
                self,
                num_entities,
                fit_entities.count(),
            )
        return getattr(self, num_entities)

    @property
    def queries_count(self) -> int:
        """
        :returns: number of queries the model was trained on
        """
        return self._get_fit_counts("query")

    @property
    def items_count(self) -> int:
        """
        :returns: number of items the model was trained on
        """
        return self._get_fit_counts("item")

    def _get_fit_dims(self, entity: str) -> int:
        dim_size = f"_{entity}_dim_size"
        fit_entities = self.fit_queries if entity == "query" else self.fit_items
        column = self.query_column if entity == "query" else self.item_column
        if not hasattr(self, dim_size):
            setattr(
                self,
                dim_size,
                fit_entities.agg({column: "max"}).first()[0] + 1,
            )
        return getattr(self, dim_size)

    @property
    def _query_dim(self) -> int:
        """
        :returns: dimension of queries matrix (maximal query id + 1)
        """
        return self._get_fit_dims("query")

    @property
    def _item_dim(self) -> int:
        """
        :returns: dimension of items matrix (maximal item idx + 1)
        """
        return self._get_fit_dims("item")

    def _fit_predict(
        self,
        dataset: Dataset,
        k: int,
        queries: Optional[Union[SparkDataFrame, Iterable]] = None,
        items: Optional[Union[SparkDataFrame, Iterable]] = None,
        filter_seen_items: bool = True,
        recs_file_path: Optional[str] = None,
    ) -> Optional[SparkDataFrame]:
        self._fit_wrap(dataset)
        return self._predict_wrap(
            dataset,
            k,
            queries,
            items,
            filter_seen_items,
            recs_file_path=recs_file_path,
        )

    def _predict_pairs_wrap(
        self,
        pairs: SparkDataFrame,
        dataset: Optional[Dataset] = None,
        recs_file_path: Optional[str] = None,
        k: Optional[int] = None,
    ) -> Optional[SparkDataFrame]:
        """
        This method
        1) converts data to spark
        2) removes cold queries and items if model does not predict them
        3) calls inner _predict_pairs method of a model

        :param pairs: query-item pairs to get rating for,
            dataframe containing``[user_idx, item_idx]``.
        :param dataset: train data
            ``[user_idx, item_idx, timestamp, rating]``.
        :param recs_file_path: save recommendations at the given absolute path as parquet file.
            If None, cached and materialized recommendations dataframe  will be returned
        :return: cached dataframe with columns ``[user_idx, item_idx, rating]``
            or None if `file_path` is provided
        """
        if dataset is not None:
            interactions, query_features, item_features, pairs = (
                convert2spark(df)
                for df in [
                    dataset.interactions,
                    dataset.query_features,
                    dataset.item_features,
                    pairs,
                ]
            )
            if set(pairs.columns) != {self.item_column, self.query_column}:
                msg = "pairs must be a dataframe with columns strictly [user_idx, item_idx]"
                raise ValueError(msg)
            pairs, interactions = self._filter_cold_for_predict(pairs, interactions, "query")
            pairs, interactions = self._filter_cold_for_predict(pairs, interactions, "item")

            dataset = Dataset(
                feature_schema=dataset.feature_schema,
                interactions=interactions,
                query_features=query_features,
                item_features=item_features,
            )

        pred = self._predict_pairs(
            pairs=pairs,
            dataset=dataset,
        )

        if k:
            pred = get_top_k(
                dataframe=pred,
                partition_by_col=sf.col(self.query_column),
                order_by_col=[
                    sf.col(self.rating_column).desc(),
                ],
                k=k,
            )

        if recs_file_path is not None:
            pred.write.parquet(path=recs_file_path, mode="overwrite")
            return None

        pred.cache().count()
        return pred

    def _predict_pairs(
        self,
        pairs: SparkDataFrame,
        dataset: Optional[Dataset] = None,
    ) -> SparkDataFrame:
        """
        Fallback method to use in case ``_predict_pairs`` is not implemented.
        Simply joins ``predict`` with given ``pairs``.
        :param pairs: query-item pairs to get rating for,
            dataframe containing``[user_idx, item_idx]``.
        :param dataset: train data
            ``[user_idx, item_idx, timestamp, rating]``.
        """
        self.logger.warning(
            "native predict_pairs is not implemented for this model. "
            "Falling back to usual predict method and filtering the results."
        )

        queries = pairs.select(self.query_column).distinct()
        items = pairs.select(self.item_column).distinct()
        k = items.count()
        pred = self._predict(
            dataset=dataset,
            k=k,
            queries=queries,
            items=items,
            filter_seen_items=False,
        )

        pred = pred.join(
            pairs.select(self.query_column, self.item_column),
            on=[self.query_column, self.item_column],
            how="inner",
        )
        return pred

    def _get_features_wrap(
        self, ids: SparkDataFrame, features: Optional[SparkDataFrame]
    ) -> Optional[tuple[SparkDataFrame, int]]:
        if self.query_column not in ids.columns and self.item_column not in ids.columns:
            msg = f"{self.query_column} or {self.item_column} missing"
            raise ValueError(msg)
        vectors, rank = self._get_features(ids, features)
        return vectors, rank

    def _get_features(
        self, ids: SparkDataFrame, features: Optional[SparkDataFrame]  # noqa: ARG002
    ) -> tuple[Optional[SparkDataFrame], Optional[int]]:
        """
        Get embeddings from model

        :param ids: id ids to get embeddings for Spark DataFrame containing user_idx or item_idx
        :param features: query or item features
        :return: SparkDataFrame with biases and embeddings, and vector size
        """

        self.logger.info(
            "get_features method is not defined for the model %s. Features will not be returned.",
            str(self),
        )
        return None, None

    def _get_nearest_items_wrap(
        self,
        items: Union[SparkDataFrame, Iterable],
        k: int,
        metric: Optional[str] = "cosine_similarity",
        candidates: Optional[Union[SparkDataFrame, Iterable]] = None,
    ) -> Optional[SparkDataFrame]:
        """
        Convert indexes and leave top-k nearest items for each item in `items`.
        """
        items = get_unique_entities(items, self.item_column)
        if candidates is not None:
            candidates = get_unique_entities(candidates, self.item_column)

        nearest_items_to_filter = self._get_nearest_items(
            items=items,
            metric=metric,
            candidates=candidates,
        )

        rel_col_name = metric if metric is not None else "similarity"
        nearest_items = get_top_k(
            dataframe=nearest_items_to_filter,
            partition_by_col=sf.col("item_idx_one"),
            order_by_col=[
                sf.col(rel_col_name).desc(),
                sf.col("item_idx_two").desc(),
            ],
            k=k,
        )

        nearest_items = nearest_items.withColumnRenamed("item_idx_two", "neighbour_item_idx")
        nearest_items = nearest_items.withColumnRenamed("item_idx_one", self.item_column)
        return nearest_items

    def _get_nearest_items(
        self,
        items: SparkDataFrame,
        metric: Optional[str] = None,
        candidates: Optional[SparkDataFrame] = None,
    ) -> Optional[SparkDataFrame]:
        msg = f"item-to-item prediction is not implemented for {self}"
        raise NotImplementedError(msg)

    def _save_model(self, path: str, additional_params: Optional[dict] = None):
        saved_params = {
            "query_column": self.query_column,
            "item_column": self.item_column,
            "rating_column": self.rating_column,
            "timestamp_column": self.timestamp_column,
        }
        if additional_params is not None:
            saved_params.update(additional_params)
        save_picklable_to_parquet(saved_params, join(path, "params.dump"))

    def _load_model(self, path: str):
        loaded_params = load_pickled_from_parquet(join(path, "params.dump"))
        for param, value in loaded_params.items():
            setattr(self, param, value)


class ItemVectorModel(BaseRecommender):
    """Parent for models generating items' vector representations"""

    can_predict_item_to_item: bool = True
    item_to_item_metrics: list[str] = [
        "euclidean_distance_sim",
        "cosine_similarity",
        "dot_product",
    ]

    @abstractmethod
    def _get_item_vectors(self) -> SparkDataFrame:
        """
        Return dataframe with items' vectors as a
            spark dataframe with columns ``[item_idx, item_vector]``
        """

    def get_nearest_items(
        self,
        items: Union[SparkDataFrame, Iterable],
        k: int,
        metric: Optional[str] = "cosine_similarity",
        candidates: Optional[Union[SparkDataFrame, Iterable]] = None,
    ) -> Optional[SparkDataFrame]:
        """
        Get k most similar items be the `metric` for each of the `items`.

        :param items: spark dataframe or list of item ids to find neighbors
        :param k: number of neighbors
        :param metric: 'euclidean_distance_sim', 'cosine_similarity', 'dot_product'
        :param candidates: spark dataframe or list of items
            to consider as similar, e.g. popular/new items. If None,
            all items presented during model training are used.
        :return: dataframe with the most similar items,
            where bigger value means greater similarity.
            spark-dataframe with columns ``[item_idx, neighbour_item_idx, similarity]``
        """
        if metric not in self.item_to_item_metrics:
            msg = f"Select one of the valid distance metrics: {self.item_to_item_metrics}"
            raise ValueError(msg)

        return self._get_nearest_items_wrap(
            items=items,
            k=k,
            metric=metric,
            candidates=candidates,
        )

    def _get_nearest_items(
        self,
        items: SparkDataFrame,
        metric: str = "cosine_similarity",
        candidates: Optional[SparkDataFrame] = None,
    ) -> SparkDataFrame:
        """
        Return distance metric value for all available close items filtered by `candidates`.

        :param items: ids to find neighbours, spark dataframe with column ``item_idx``
        :param metric: 'euclidean_distance_sim' calculated as 1/(1 + euclidean_distance),
            'cosine_similarity', 'dot_product'
        :param candidates: items among which we are looking for similar,
            e.g. popular/new items. If None, all items presented during model training are used.
        :return: dataframe with neighbours,
            spark-dataframe with columns ``[item_idx_one, item_idx_two, similarity]``
        """
        dist_function = cosine_similarity
        if metric == "euclidean_distance_sim":
            dist_function = vector_euclidean_distance_similarity
        elif metric == "dot_product":
            dist_function = vector_dot

        items_vectors = self._get_item_vectors()
        left_part = (
            items_vectors.withColumnRenamed(self.item_column, "item_idx_one")
            .withColumnRenamed("item_vector", "item_vector_one")
            .join(
                items.select(sf.col(self.item_column).alias("item_idx_one")),
                on="item_idx_one",
            )
        )

        right_part = items_vectors.withColumnRenamed(self.item_column, "item_idx_two").withColumnRenamed(
            "item_vector", "item_vector_two"
        )

        if candidates is not None:
            right_part = right_part.join(
                candidates.withColumnRenamed(self.item_column, "item_idx_two"),
                on="item_idx_two",
            )

        joined_factors = left_part.join(right_part, on=sf.col("item_idx_one") != sf.col("item_idx_two"))

        joined_factors = joined_factors.withColumn(
            metric,
            dist_function(sf.col("item_vector_one"), sf.col("item_vector_two")),
        )

        similarity_matrix = joined_factors.select("item_idx_one", "item_idx_two", metric)

        return similarity_matrix


class HybridRecommender(BaseRecommender, ABC):
    """Base class for models that can use extra features"""

    def fit(
        self,
        dataset: Dataset,
    ) -> None:
        """
        Fit a recommendation model

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :return:
        """
        self._fit_wrap(dataset=dataset)

    def predict(
        self,
        dataset: Dataset,
        k: int,
        queries: Optional[Union[SparkDataFrame, Iterable]] = None,
        items: Optional[Union[SparkDataFrame, Iterable]] = None,
        filter_seen_items: bool = True,
        recs_file_path: Optional[str] = None,
    ) -> Optional[SparkDataFrame]:
        """
        Get recommendations

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :param k: number of recommendations for each query
        :param queries: queries to create recommendations for
            dataframe containing ``[user_idx]`` or ``array-like``;
            if ``None``, recommend to all queries from ``interactions``
        :param items: candidate items for recommendations
            dataframe containing ``[item_idx]`` or ``array-like``;
            if ``None``, take all items from ``interactions``.
            If it contains new items, ``rating`` for them will be ``0``.
        :param filter_seen_items: flag to remove seen items from recommendations based on ``interactions``.
        :param recs_file_path: save recommendations at the given absolute path as parquet file.
            If None, cached and materialized recommendations dataframe  will be returned
        :return: cached recommendation dataframe with columns ``[user_idx, item_idx, rating]``
            or None if `file_path` is provided

        """
        return self._predict_wrap(
            dataset=dataset,
            k=k,
            queries=queries,
            items=items,
            filter_seen_items=filter_seen_items,
            recs_file_path=recs_file_path,
        )

    def fit_predict(
        self,
        dataset: Dataset,
        k: int,
        queries: Optional[Union[SparkDataFrame, Iterable]] = None,
        items: Optional[Union[SparkDataFrame, Iterable]] = None,
        filter_seen_items: bool = True,
        recs_file_path: Optional[str] = None,
    ) -> Optional[SparkDataFrame]:
        """
        Fit model and get recommendations

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :param k: number of recommendations for each query
        :param queries: queries to create recommendations for
            dataframe containing ``[user_idx]`` or ``array-like``;
            if ``None``, recommend to all queries from ``interactions``
        :param items: candidate items for recommendations
            dataframe containing ``[item_idx]`` or ``array-like``;
            if ``None``, take all items from ``interactions``.
            If it contains new items, ``rating`` for them will be ``0``.
        :param filter_seen_items: flag to remove seen items from recommendations based on ``interactions``.
        :param recs_file_path: save recommendations at the given absolute path as parquet file.
            If None, cached and materialized recommendations dataframe  will be returned
        :return: cached recommendation dataframe with columns ``[user_idx, item_idx, rating]``
            or None if `file_path` is provided
        """
        return self._fit_predict(
            dataset=dataset,
            k=k,
            queries=queries,
            items=items,
            filter_seen_items=filter_seen_items,
            recs_file_path=recs_file_path,
        )

    def predict_pairs(
        self,
        pairs: SparkDataFrame,
        dataset: Optional[Dataset] = None,
        recs_file_path: Optional[str] = None,
        k: Optional[int] = None,
    ) -> Optional[SparkDataFrame]:
        """
        Get recommendations for specific query-item ``pairs``.
        If a model can't produce recommendation
        for specific pair it is removed from the resulting dataframe.

        :param pairs: dataframe with pairs to calculate rating for, ``[user_idx, item_idx]``.
        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :param recs_file_path: save recommendations at the given absolute path as parquet file.
            If None, cached and materialized recommendations dataframe  will be returned
        :param k: top-k items for each query from pairs.
        :return: cached recommendation dataframe with columns ``[user_idx, item_idx, rating]``
            or None if `file_path` is provided
        """
        return self._predict_pairs_wrap(
            pairs=pairs,
            dataset=dataset,
            recs_file_path=recs_file_path,
            k=k,
        )

    def get_features(
        self, ids: SparkDataFrame, features: Optional[SparkDataFrame]
    ) -> Optional[tuple[SparkDataFrame, int]]:
        """
        Returns query or item feature vectors as a Column with type ArrayType
        If a model does not have a vector for some ids they are not present in the final result.

        :param ids: Spark DataFrame with unique ids
        :param features: Spark DataFrame with features for provided ids
        :return: feature vectors
        """
        return self._get_features_wrap(ids, features)


class Recommender(BaseRecommender, ABC):
    """Usual recommender class for models without features."""

    def fit(self, dataset: Dataset) -> None:
        """
        Fit a recommendation model

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :return:
        """
        self._fit_wrap(dataset=dataset)

    def predict(
        self,
        dataset: Dataset,
        k: int,
        queries: Optional[Union[SparkDataFrame, Iterable]] = None,
        items: Optional[Union[SparkDataFrame, Iterable]] = None,
        filter_seen_items: bool = True,
        recs_file_path: Optional[str] = None,
    ) -> Optional[SparkDataFrame]:
        """
        Get recommendations

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :param k: number of recommendations for each query
        :param queries: queries to create recommendations for
            dataframe containing ``[user_idx]`` or ``array-like``;
            if ``None``, recommend to all queries from ``interactions``
        :param items: candidate items for recommendations
            dataframe containing ``[item_idx]`` or ``array-like``;
            if ``None``, take all items from ``interactions``.
            If it contains new items, ``rating`` for them will be ``0``.
        :param filter_seen_items: flag to remove seen items from recommendations based on ``interactions``.
        :param recs_file_path: save recommendations at the given absolute path as parquet file.
            If None, cached and materialized recommendations dataframe  will be returned
        :return: cached recommendation dataframe with columns ``[user_idx, item_idx, rating]``
            or None if `file_path` is provided
        """
        return self._predict_wrap(
            dataset=dataset,
            k=k,
            queries=queries,
            items=items,
            filter_seen_items=filter_seen_items,
            recs_file_path=recs_file_path,
        )

    def predict_pairs(
        self,
        pairs: SparkDataFrame,
        dataset: Optional[Dataset] = None,
        recs_file_path: Optional[str] = None,
        k: Optional[int] = None,
    ) -> Optional[SparkDataFrame]:
        """
        Get recommendations for specific query-item ``pairs``.
        If a model can't produce recommendation
        for specific pair it is removed from the resulting dataframe.

        :param pairs: dataframe with pairs to calculate rating for, ``[user_idx, item_idx]``.
        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :param recs_file_path: save recommendations at the given absolute path as parquet file.
            If None, cached and materialized recommendations dataframe  will be returned
        :param k: top-k items for each query from pairs.
        :return: cached recommendation dataframe with columns ``[user_idx, item_idx, rating]``
            or None if `file_path` is provided
        """
        return self._predict_pairs_wrap(
            pairs=pairs,
            dataset=dataset,
            recs_file_path=recs_file_path,
            k=k,
        )

    def fit_predict(
        self,
        dataset: Dataset,
        k: int,
        queries: Optional[Union[SparkDataFrame, Iterable]] = None,
        items: Optional[Union[SparkDataFrame, Iterable]] = None,
        filter_seen_items: bool = True,
        recs_file_path: Optional[str] = None,
    ) -> Optional[SparkDataFrame]:
        """
        Fit model and get recommendations

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :param k: number of recommendations for each query
        :param queries: queries to create recommendations for
            dataframe containing ``[user_idx]`` or ``array-like``;
            if ``None``, recommend to all queries from ``interactions``
        :param items: candidate items for recommendations
            dataframe containing ``[item_idx]`` or ``array-like``;
            if ``None``, take all items from ``interactions``.
            If it contains new items, ``rating`` for them will be ``0``.
        :param filter_seen_items: flag to remove seen items from recommendations based on ``interactions``.
        :param recs_file_path: save recommendations at the given absolute path as parquet file.
            If None, cached and materialized recommendations dataframe  will be returned
        :return: cached recommendation dataframe with columns ``[user_idx, item_idx, rating]``
            or None if `file_path` is provided
        """
        return self._fit_predict(
            dataset=dataset,
            k=k,
            queries=queries,
            items=items,
            filter_seen_items=filter_seen_items,
            recs_file_path=recs_file_path,
        )

    def get_features(self, ids: SparkDataFrame) -> Optional[tuple[SparkDataFrame, int]]:
        """
        Returns query or item feature vectors as a Column with type ArrayType

        :param ids: Spark DataFrame with unique ids
        :return: feature vectors.
            If a model does not have a vector for some ids they are not present in the final result.
        """
        return self._get_features_wrap(ids, None)


class QueryRecommender(BaseRecommender, ABC):
    """Base class for models that use query features
    but not item features. ``interactions`` is not required for this class."""

    def fit(
        self,
        dataset: Dataset,
    ) -> None:
        """
        Finds query clusters and calculates item similarity in that clusters.

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :return:
        """
        self._fit_wrap(dataset=dataset)

    def predict(
        self,
        dataset: Dataset,
        k: int,
        queries: Optional[Union[SparkDataFrame, Iterable]] = None,
        items: Optional[Union[SparkDataFrame, Iterable]] = None,
        filter_seen_items: bool = True,
        recs_file_path: Optional[str] = None,
    ) -> Optional[SparkDataFrame]:
        """
        Get recommendations

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :param k: number of recommendations for each query
        :param queries: queries to create recommendations for
            dataframe containing ``[user_idx]`` or ``array-like``;
            if ``None``, recommend to all queries from ``interactions``
        :param items: candidate items for recommendations
            dataframe containing ``[item_idx]`` or ``array-like``;
            if ``None``, take all items from ``interactions``.
            If it contains new items, ``rating`` for them will be ``0``.
        :param filter_seen_items: flag to remove seen items from recommendations based on ``interactions``.
        :param recs_file_path: save recommendations at the given absolute path as parquet file.
            If None, cached and materialized recommendations dataframe  will be returned
        :return: cached recommendation dataframe with columns ``[user_idx, item_idx, rating]``
            or None if `file_path` is provided
        """
        if not dataset or not dataset.query_features:
            msg = "Query features are missing for predict"
            raise ValueError(msg)

        return self._predict_wrap(
            dataset=dataset,
            k=k,
            filter_seen_items=filter_seen_items,
            queries=queries,
            items=items,
            recs_file_path=recs_file_path,
        )

    def predict_pairs(
        self,
        pairs: SparkDataFrame,
        dataset: Dataset,
        recs_file_path: Optional[str] = None,
        k: Optional[int] = None,
    ) -> Optional[SparkDataFrame]:
        """
        Get recommendations for specific query-item ``pairs``.
        If a model can't produce recommendation
        for specific pair it is removed from the resulting dataframe.

        :param pairs: dataframe with pairs to calculate rating for, ``[user_idx, item_idx]``.
        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :param recs_file_path: save recommendations at the given absolute path as parquet file.
            If None, cached and materialized recommendations dataframe  will be returned
        :param k: top-k items for each query from pairs.
        :return: cached recommendation dataframe with columns ``[user_idx, item_idx, rating]``
            or None if `file_path` is provided
        """
        if not dataset or not dataset.query_features:
            msg = "Query features are missing for predict"
            raise ValueError(msg)

        return self._predict_pairs_wrap(
            pairs=pairs,
            dataset=dataset,
            recs_file_path=recs_file_path,
            k=k,
        )


class NonPersonalizedRecommender(Recommender, ABC):
    """Base class for non-personalized recommenders with popularity statistics."""

    can_predict_cold_queries = True
    can_predict_cold_items = True
    item_popularity: SparkDataFrame
    add_cold_items: bool
    cold_weight: float
    sample: bool
    fill: float
    seed: Optional[int] = None

    def __init__(self, add_cold_items: bool, cold_weight: float):
        self.add_cold_items = add_cold_items
        if 0 < cold_weight <= 1:
            self.cold_weight = cold_weight
        else:
            msg = "`cold_weight` value should be in interval (0, 1]"
            raise ValueError(msg)

    @property
    def _dataframes(self):
        return {"item_popularity": self.item_popularity}

    def _save_model(self, path: str, additional_params: Optional[dict] = None):  # noqa: ARG002
        super()._save_model(path, additional_params={"fill": self.fill})

    def _clear_cache(self):
        if hasattr(self, "item_popularity"):
            self.item_popularity.unpersist()

    @staticmethod
    def _calc_fill(item_popularity: SparkDataFrame, weight: float, rating_column: str) -> float:
        """
        Calculating a fill value a the minimal rating
        calculated during model training multiplied by weight.
        """
        return item_popularity.select(sf.min(rating_column)).first()[0] * weight

    @staticmethod
    def _check_rating(dataset: Dataset):
        rating_column = dataset.feature_schema.interactions_rating_column
        vals = dataset.interactions.select(rating_column).where(
            (sf.col(rating_column) != 1) & (sf.col(rating_column) != 0)
        )
        if vals.count() > 0:
            msg = "Rating values in interactions must be 0 or 1"
            raise ValueError(msg)

    def _get_selected_item_popularity(self, items: SparkDataFrame) -> SparkDataFrame:
        """
        Choose only required item from `item_popularity` dataframe
        for further recommendations generation.
        """
        return self.item_popularity.join(
            items,
            on=self.item_column,
            how="right" if self.add_cold_items else "inner",
        ).fillna(value=self.fill, subset=[self.rating_column])

    @staticmethod
    def _calc_max_hist_len(dataset: Dataset, queries: SparkDataFrame) -> int:
        query_column = dataset.feature_schema.query_id_column
        item_column = dataset.feature_schema.item_id_column
        max_hist_len = (
            (
                dataset.interactions.join(queries, on=query_column)
                .groupBy(query_column)
                .agg(sf.countDistinct(item_column).alias("items_count"))
            )
            .select(sf.max("items_count"))
            .first()[0]
        )
        # all queries have empty history
        if max_hist_len is None:
            max_hist_len = 0

        return max_hist_len

    def _predict_without_sampling(
        self,
        dataset: Dataset,
        k: int,
        queries: SparkDataFrame,
        items: SparkDataFrame,
        filter_seen_items: bool = True,
    ) -> SparkDataFrame:
        """
        Regular prediction for popularity-based models,
        top-k most relevant items from `items` are chosen for each query
        """
        selected_item_popularity = self._get_selected_item_popularity(items)
        selected_item_popularity = selected_item_popularity.withColumn(
            "rank",
            sf.row_number().over(Window.orderBy(sf.col(self.rating_column).desc(), sf.col(self.item_column).desc())),
        )

        if filter_seen_items and dataset is not None:
            query_to_num_items = (
                dataset.interactions.join(queries, on=self.query_column)
                .groupBy(self.query_column)
                .agg(sf.countDistinct(self.item_column).alias("num_items"))
            )
            queries = queries.join(query_to_num_items, on=self.query_column, how="left")
            queries = queries.fillna(0, "num_items")
            # 'selected_item_popularity' truncation by k + max_seen
            max_seen = queries.select(sf.coalesce(sf.max("num_items"), sf.lit(0))).first()[0]
            selected_item_popularity = selected_item_popularity.filter(sf.col("rank") <= k + max_seen)
            return queries.join(
                selected_item_popularity,
                on=(sf.col("rank") <= k + sf.col("num_items")),
                how="left",
            )

        return queries.crossJoin(selected_item_popularity.filter(sf.col("rank") <= k)).drop("rank")

    def get_items_pd(self, items: SparkDataFrame) -> pd.DataFrame:
        """
        Function to calculate normalized popularities(in fact, probabilities)
        of given items. Returns pandas DataFrame.
        """
        selected_item_popularity = self._get_selected_item_popularity(items)
        selected_item_popularity = selected_item_popularity.withColumn(
            self.rating_column,
            sf.when(sf.col(self.rating_column) == sf.lit(0.0), 0.1**6).otherwise(sf.col(self.rating_column)),
        )

        warnings.warn(
            "Prediction with sampling performs spark to pandas convertion to master node, "
            "this may lead to OOM exception for large item catalogue.",
            SparkCollectToMasterWarning,
        )

        items_pd = selected_item_popularity.withColumn(
            "probability",
            sf.col(self.rating_column) / selected_item_popularity.select(sf.sum(self.rating_column)).first()[0],
        ).toPandas()

        return items_pd

    def _predict_with_sampling(
        self,
        dataset: Dataset,
        k: int,
        queries: SparkDataFrame,
        items: SparkDataFrame,
        filter_seen_items: bool = True,
    ) -> SparkDataFrame:
        """
        Randomized prediction for popularity-based models,
        top-k items from `items` are sampled for each query based with
        probability proportional to items' popularity
        """
        items_pd = self.get_items_pd(items)

        rec_schema = get_schema(
            query_column=self.query_column,
            item_column=self.item_column,
            rating_column=self.rating_column,
            has_timestamp=False,
        )

        if items_pd.shape[0] == 0:
            return State().session.createDataFrame([], rec_schema)

        seed = self.seed
        query_column = self.query_column
        item_column = self.item_column
        rating_column = self.rating_column
        class_name = self.__class__.__name__

        def grouped_map(
            pandas_df: PandasDataFrame,
        ) -> PandasDataFrame:  # pragma: no cover
            query_idx = pandas_df[query_column][0]
            cnt = pandas_df["cnt"][0]

            local_rng = default_rng(seed + query_idx) if seed is not None else default_rng()

            items_positions = local_rng.choice(
                np.arange(items_pd.shape[0]),
                size=cnt,
                p=items_pd["probability"].values,
                replace=False,
            )

            # workaround to unify RandomRec and UCB
            if class_name == "RandomRec":
                return PandasDataFrame(
                    {
                        query_column: cnt * [query_idx],
                        item_column: items_pd[item_column].sort_values().values[items_positions],
                        rating_column: local_rng.uniform(size=items_pd.shape[0])[items_positions],
                    }
                )
            else:
                rating = items_pd["probability"].values[items_positions]

            return PandasDataFrame(
                {
                    query_column: cnt * [query_idx],
                    item_column: items_pd[item_column].values[items_positions],
                    rating_column: rating,
                }
            )

        if dataset is not None and filter_seen_items:
            recs = (
                dataset.interactions.select(self.query_column, self.item_column)
                .distinct()
                .join(queries, how="right", on=self.query_column)
                .groupby(self.query_column)
                .agg(sf.countDistinct(self.item_column).alias("cnt"))
                .selectExpr(
                    self.query_column,
                    f"LEAST(cnt + {k}, {items_pd.shape[0]}) AS cnt",
                )
            )
        else:
            recs = queries.withColumn("cnt", sf.lit(min(k, items_pd.shape[0])))

        return recs.groupby(self.query_column).applyInPandas(grouped_map, rec_schema)

    def _predict(
        self,
        dataset: Dataset,
        k: int,
        queries: SparkDataFrame,
        items: SparkDataFrame,
        filter_seen_items: bool = True,
    ) -> SparkDataFrame:
        if self.sample:
            return self._predict_with_sampling(
                dataset=dataset,
                k=k,
                queries=queries,
                items=items,
                filter_seen_items=filter_seen_items,
            )
        else:
            return self._predict_without_sampling(dataset, k, queries, items, filter_seen_items)

    def _predict_pairs(
        self,
        pairs: SparkDataFrame,
        dataset: Optional[Dataset] = None,  # noqa: ARG002
    ) -> SparkDataFrame:
        return (
            pairs.join(
                self.item_popularity,
                on=self.item_column,
                how="left" if self.add_cold_items else "inner",
            )
            .fillna(value=self.fill, subset=[self.rating_column])
            .select(self.query_column, self.item_column, self.rating_column)
        )

    def _predict_proba(
        self,
        dataset: Dataset,
        k: int,
        queries: SparkDataFrame,
        items: SparkDataFrame,
        filter_seen_items: bool = True,
    ) -> np.ndarray:
        """
        Inner method where model actually predicts probability estimates.

        Mainly used in ```OBPOfflinePolicyLearner```.

        :param dataset: historical interactions with query/item features
            ``[user_idx, item_idx, timestamp, rating]``
        :param k: number of recommendations for each user
        :param queries: queries to create recommendations for
            dataframe containing ``[user_idx]`` or ``array-like``;
            if ``None``, recommend to all queries from ``interactions``
        :param items: candidate items for recommendations
            dataframe containing ``[item_idx]`` or ``array-like``;
            if ``None``, take all items from ``interactions``.
            If it contains new items, ``rating`` for them will be ``0``.
        :param filter_seen_items: flag to remove seen items from recommendations based on ``interactions``.
        :return: distribution over items for each user with shape
            ``(n_users, n_items, k)``
            where we have probability for each user to choose item at fixed position(top-k).
        """

        n_users = queries.select("user_idx").count()
        n_items = items.select("item_idx").count()

        if self.sample:
            items_pd = self.get_items_pd(items)

            items_idx = items_pd["item_idx"].to_numpy()
            items_idx_inv = np.zeros_like(items_idx)
            items_idx_inv[items_idx] = np.arange(len(items_idx))

            items_pd = items_pd["probability"].to_numpy()[items_idx_inv]

            return np.tile(items_pd, (n_users, k)).reshape(n_users, k, n_items).transpose((0, 2, 1))

        return super()._predict_proba(dataset, k, queries, items, filter_seen_items)



from .base_metric import Metric


class MAP(Metric):
    """
    Mean Average Precision -- average the ``Precision`` at relevant positions \
        for each user, and then calculate the mean across all users.

    .. math::
        &AP@K(i) = \\frac {1}{\\min(K, |Rel_i|)} \\sum_{j=1}^{K}\\mathbb{1}_{r_{ij}}Precision@j(i)

        &MAP@K = \\frac {\\sum_{i=1}^{N}AP@K(i)}{N}

    :math:`\\mathbb{1}_{r_{ij}}` -- indicator function showing if user :math:`i` interacted with item :math:`j`

    :math:`|Rel_i|` -- the number of relevant items for user :math:`i`

    >>> recommendations
       query_id  item_id  rating
    0         1        3    0.6
    1         1        7    0.5
    2         1       10    0.4
    3         1       11    0.3
    4         1        2    0.2
    5         2        5    0.6
    6         2        8    0.5
    7         2       11    0.4
    8         2        1    0.3
    9         2        3    0.2
    10        3        4    1.0
    11        3        9    0.5
    12        3        2    0.1
    >>> groundtruth
       query_id  item_id
    0         1        5
    1         1        6
    2         1        7
    3         1        8
    4         1        9
    5         1       10
    6         2        6
    7         2        7
    8         2        4
    9         2       10
    10        2       11
    11        3        1
    12        3        2
    13        3        3
    14        3        4
    15        3        5
    >>> from replay.metrics import Median, ConfidenceInterval, PerUser
    >>> MAP(2)(recommendations, groundtruth)
    {'MAP@2': 0.25}
    >>> MAP(2, mode=PerUser())(recommendations, groundtruth)
    {'MAP-PerUser@2': {1: 0.25, 2: 0.0, 3: 0.5}}
    >>> MAP(2, mode=Median())(recommendations, groundtruth)
    {'MAP-Median@2': 0.25}
    >>> MAP(2, mode=ConfidenceInterval(alpha=0.95))(recommendations, groundtruth)
    {'MAP-ConfidenceInterval@2': 0.282896433519043}
    <BLANKLINE>
    """

    @staticmethod
    def _get_metric_value_by_user(ks: list[int], ground_truth: list, pred: list) -> list[float]:
        if not ground_truth or not pred:
            return [0.0 for _ in ks]
        res = []
        for k in ks:
            length = min(k, len(pred))
            max_good = min(k, len(ground_truth))
            tp_cum: int = 0
            result: float = 0.0
            for i in range(length):
                if pred[i] in ground_truth:
                    tp_cum += 1
                    result += tp_cum / (i + 1)
            res.append(result / max_good)
        return res



"""
Painless creation and retrieval of Spark sessions
"""

import logging
import os
import sys
from math import floor
from typing import Any, Optional

import psutil

from .types import PYSPARK_AVAILABLE, MissingImport

if PYSPARK_AVAILABLE:
    from pyspark import __version__ as pyspark_version
    from pyspark.sql import SparkSession
else:
    SparkSession = MissingImport


def get_spark_session(
    spark_memory: Optional[int] = None,
    shuffle_partitions: Optional[int] = None,
    core_count: Optional[int] = None,
) -> SparkSession:
    """
    Get default SparkSession

    :param spark_memory: GB of memory allocated for Spark;
        70% of RAM by default.
    :param shuffle_partitions: number of partitions for Spark; triple CPU count by default
    :param core_count: Count of cores to execute, ``-1`` means using all available cores.
        If ``None`` then checking out environment variable ``REPLAY_SPARK_CORE_COUNT``,
        if variable is not set then using ``-1``.
        Default: ``None``.
    """
    if os.environ.get("SCRIPT_ENV", None) == "cluster":  # pragma: no cover
        return SparkSession.builder.getOrCreate()

    os.environ["PYSPARK_PYTHON"] = sys.executable
    os.environ["PYSPARK_DRIVER_PYTHON"] = sys.executable

    if os.environ.get("REPLAY_JAR_PATH"):  # pragma: no cover
        path_to_replay_jar = os.environ.get("REPLAY_JAR_PATH")
    else:
        if pyspark_version.startswith("3.1"):  # pragma: no cover
            path_to_replay_jar = (
                "https://repo1.maven.org/maven2/io/github/sb-ai-lab/replay_2.12/3.1.3/replay_2.12-3.1.3.jar"
            )
        elif pyspark_version.startswith(("3.2", "3.3")):  # pragma: no cover
            path_to_replay_jar = "https://repo1.maven.org/maven2/io/github/sb-ai-lab/replay_2.12/3.2.0_als_metrics/replay_2.12-3.2.0_als_metrics.jar"
        elif pyspark_version.startswith("3.4"):  # pragma: no cover
            path_to_replay_jar = "https://repo1.maven.org/maven2/io/github/sb-ai-lab/replay_after_fix_2.12/0.1/replay_after_fix_2.12-0.1.jar"
        else:  # pragma: no cover
            path_to_replay_jar = (
                "https://repo1.maven.org/maven2/io/github/sb-ai-lab/replay_2.12/3.1.3/replay_2.12-3.1.3.jar"
            )
            logging.warning(
                "Replay ALS model support only spark 3.1-3.4 versions! Replay will use "
                "'https://repo1.maven.org/maven2/io/github/sb-ai-lab/replay_2.12/3.1.3/replay_2.12-3.1.3.jar' "
                "in 'spark.jars' property."
            )

    if core_count is None:  # checking out env variable
        core_count = int(os.environ.get("REPLAY_SPARK_CORE_COUNT", "-1"))
    if spark_memory is None:
        env_var = os.environ.get("REPLAY_SPARK_MEMORY")
        spark_memory = int(env_var) if env_var is not None else floor(psutil.virtual_memory().total / 1024**3 * 0.7)
    if shuffle_partitions is None:
        shuffle_partitions = os.cpu_count() * 3
    driver_memory = f"{spark_memory}g"
    user_home = os.environ["HOME"]
    spark_session_builder = (
        SparkSession.builder.config("spark.driver.memory", driver_memory)
        .config(
            "spark.driver.extraJavaOptions",
            "-Dio.netty.tryReflectionSetAccessible=true",
        )
        .config("spark.jars", path_to_replay_jar)
        .config("spark.sql.shuffle.partitions", str(shuffle_partitions))
        .config("spark.local.dir", os.path.join(user_home, "tmp"))
        .config("spark.driver.maxResultSize", "4g")
        .config("spark.driver.bindAddress", "127.0.0.1")
        .config("spark.driver.host", "localhost")
        .config("spark.sql.execution.arrow.enabled", "true")
        .config("spark.kryoserializer.buffer.max", "256m")
        .config("spark.files.overwrite", "true")
        .master(f"local[{'*' if core_count == -1 else core_count}]")
    )

    return spark_session_builder.getOrCreate()


def logger_with_settings() -> logging.Logger:
    """Set up default logging"""
    spark_logger = logging.getLogger("py4j")
    spark_logger.setLevel(logging.WARN)
    logger = logging.getLogger("replay")
    formatter = logging.Formatter(
        "%(asctime)s, %(name)s, %(levelname)s: %(message)s",
        datefmt="%d-%b-%y %H:%M:%S",
    )
    hdlr = logging.StreamHandler()
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)
    logger.setLevel(logging.INFO)
    return logger


class Borg:
    """
    This class allows to share objects between instances.
    """

    _shared_state: dict[str, Any] = {}

    def __init__(self):
        self.__dict__ = self._shared_state


class State(Borg):
    """
    All modules look for Spark session via this class. You can put your own session here.
    """

    def __init__(
        self,
        session: Optional[SparkSession] = None,
    ):
        Borg.__init__(self)
        if not hasattr(self, "logger_set"):
            self.logger = logger_with_settings()
            self.logger_set = True

        if session is None:
            if not hasattr(self, "session"):
                self.session = get_spark_session()
        else:
            self.session = session



from typing import Optional

import polars as pl

from replay.utils import PYSPARK_AVAILABLE, DataFrameLike, PandasDataFrame, PolarsDataFrame, SparkDataFrame

from .base_splitter import Splitter

if PYSPARK_AVAILABLE:
    import pyspark.sql.functions as sf
    from pyspark.sql import Window
    from pyspark.sql.types import IntegerType


class RatioSplitter(Splitter):
    """
    Split interactions into train and test by ratio. Split is made for each user separately.

    >>> from datetime import datetime
    >>> import pandas as pd
    >>> columns = ["query_id", "item_id", "timestamp"]
    >>> data = [
    ...     (1, 1, "01-01-2020"),
    ...     (1, 2, "02-01-2020"),
    ...     (1, 3, "03-01-2020"),
    ...     (1, 4, "04-01-2020"),
    ...     (1, 5, "05-01-2020"),
    ...     (2, 1, "06-01-2020"),
    ...     (2, 2, "07-01-2020"),
    ...     (2, 3, "08-01-2020"),
    ...     (2, 9, "09-01-2020"),
    ...     (2, 10, "10-01-2020"),
    ...     (3, 1, "01-01-2020"),
    ...     (3, 5, "02-01-2020"),
    ...     (3, 3, "03-01-2020"),
    ...     (3, 1, "04-01-2020"),
    ...     (3, 2, "05-01-2020"),
    ... ]
    >>> dataset = pd.DataFrame(data, columns=columns)
    >>> dataset["timestamp"] = pd.to_datetime(dataset["timestamp"], format="%d-%m-%Y")
    >>> dataset
        query_id  item_id  timestamp
    0         1        1 2020-01-01
    1         1        2 2020-01-02
    2         1        3 2020-01-03
    3         1        4 2020-01-04
    4         1        5 2020-01-05
    5         2        1 2020-01-06
    6         2        2 2020-01-07
    7         2        3 2020-01-08
    8         2        9 2020-01-09
    9         2       10 2020-01-10
    10        3        1 2020-01-01
    11        3        5 2020-01-02
    12        3        3 2020-01-03
    13        3        1 2020-01-04
    14        3        2 2020-01-05
    >>> splitter = RatioSplitter(
    ...     test_size=0.5,
    ...     divide_column="query_id",
    ...     query_column="query_id",
    ...     item_column="item_id"
    ... )
    >>> train, test = splitter.split(dataset)
    >>> train
       query_id  item_id  timestamp
    0         1        1 2020-01-01
    1         1        2 2020-01-02
    5         2        1 2020-01-06
    6         2        2 2020-01-07
    10        3        1 2020-01-01
    11        3        5 2020-01-02
    >>> test
       query_id  item_id  timestamp
    2         1        3 2020-01-03
    3         1        4 2020-01-04
    4         1        5 2020-01-05
    7         2        3 2020-01-08
    8         2        9 2020-01-09
    9         2       10 2020-01-10
    12        3        3 2020-01-03
    13        3        1 2020-01-04
    14        3        2 2020-01-05
    <BLANKLINE>
    """

    _init_arg_names = [
        "test_size",
        "divide_column",
        "drop_cold_users",
        "drop_cold_items",
        "query_column",
        "item_column",
        "timestamp_column",
        "min_interactions_per_group",
        "split_by_fractions",
        "session_id_column",
        "session_id_processing_strategy",
    ]

    def __init__(
        self,
        test_size: float,
        divide_column: str = "query_id",
        drop_cold_users: bool = False,
        drop_cold_items: bool = False,
        query_column: str = "query_id",
        item_column: str = "item_id",
        timestamp_column: str = "timestamp",
        min_interactions_per_group: Optional[int] = None,
        split_by_fractions: bool = True,
        session_id_column: Optional[str] = None,
        session_id_processing_strategy: str = "test",
    ):
        """
        :param ratio: test size, must be in :math:`(0, 1)`.
        :param divide_column: Name of column for dividing
            in dataframe, default: ``query_id``.
        :param drop_cold_users: Drop users from test DataFrame.
            which are not in train DataFrame, default: False.
        :param drop_cold_items: Drop items from test DataFrame
            which are not in train DataFrame, default: False.
        :param query_column: Name of query interaction column.
            If ``drop_cold_users`` is ``False``, then you can omit this parameter.
            Default: ``query_id``.
        :param item_column: Name of item interaction column.
            If ``drop_cold_items`` is ``False``, then you can omit this parameter.
            Default: ``item_id``.
        :param timestamp_column: Name of time column,
            Default: ``timestamp``.
        :param min_interactions_per_group: minimal required interactions per group to make first split.
            if value is less than min_interactions_per_group, than whole group goes to train.
            If not set, than any amount of interactions will be split.
            default: ``None``.
        :param split_by_fractions: the variable that is responsible for using the split by fractions.
            Split by fractions means that each line is marked with its fraq (line number / number of lines)
            and only those lines with a fraq > test_ratio get into the test.
            Split not by fractions means that the number of rows in the train is calculated by rounding the formula:
            the total number of rows minus the number of rows multiplied by the test ratio.
            The difference between these two methods is that due to rounding in the second method,
            1 more interaction in each group (1 item for each user) falls into the train.
            When split by fractions, these items fall into the test.
            default: ``True``.
        :param session_id_column: Name of session id column, which values can not be split,
            default: ``None``.
        :param session_id_processing_strategy: strategy of processing session if it is split,
            values: ``train, test``, train: whole split session goes to train. test: same but to test.
            default: ``test``.
        """
        super().__init__(
            drop_cold_users=drop_cold_users,
            drop_cold_items=drop_cold_items,
            query_column=query_column,
            item_column=item_column,
            timestamp_column=timestamp_column,
            session_id_column=session_id_column,
            session_id_processing_strategy=session_id_processing_strategy,
        )
        self.divide_column = divide_column
        self._precision = 3
        self.min_interactions_per_group = min_interactions_per_group
        self.split_by_fractions = split_by_fractions
        if test_size < 0 or test_size > 1:
            msg = "test_size must between 0 and 1"
            raise ValueError(msg)
        self.test_size = test_size

    def _add_time_partition(self, interactions: DataFrameLike) -> DataFrameLike:
        if isinstance(interactions, SparkDataFrame):
            return self._add_time_partition_to_spark(interactions)
        if isinstance(interactions, PandasDataFrame):
            return self._add_time_partition_to_pandas(interactions)
        if isinstance(interactions, PolarsDataFrame):
            return self._add_time_partition_to_polars(interactions)

        msg = f"{self} is not implemented for {type(interactions)}"
        raise NotImplementedError(msg)

    def _add_time_partition_to_pandas(self, interactions: PandasDataFrame) -> PandasDataFrame:
        res = interactions.copy(deep=True)
        res.sort_values(by=[self.divide_column, self.timestamp_column], inplace=True)
        res["row_num"] = res.groupby(self.divide_column, sort=False).cumcount() + 1

        return res

    def _add_time_partition_to_spark(self, interactions: SparkDataFrame) -> SparkDataFrame:
        res = interactions.withColumn(
            "row_num",
            sf.row_number().over(Window.partitionBy(self.divide_column).orderBy(sf.col(self.timestamp_column))),
        )

        return res

    def _add_time_partition_to_polars(self, interactions: PolarsDataFrame) -> PolarsDataFrame:
        res = interactions.sort(self.timestamp_column).with_columns(
            pl.cum_count(self.divide_column).over(self.divide_column).alias("row_num")
        )

        return res

    def _partial_split_fractions(
        self, interactions: DataFrameLike, ratio: float
    ) -> tuple[DataFrameLike, DataFrameLike]:
        res = self._add_time_partition(interactions)
        train_size = round(1 - ratio, self._precision)

        if isinstance(res, SparkDataFrame):
            return self._partial_split_fractions_spark(res, train_size)
        if isinstance(res, PandasDataFrame):
            return self._partial_split_fractions_pandas(res, train_size)
        return self._partial_split_fractions_polars(res, train_size)

    def _partial_split_fractions_pandas(
        self, interactions: PandasDataFrame, train_size: float
    ) -> tuple[PandasDataFrame, PandasDataFrame]:
        interactions["count"] = interactions.groupby(self.divide_column, sort=False)[self.divide_column].transform(len)
        interactions["frac"] = (interactions["row_num"] / interactions["count"]).round(self._precision)
        if self.min_interactions_per_group is not None:
            interactions["frac"].where(interactions["count"] >= self.min_interactions_per_group, 0, inplace=True)

        interactions["is_test"] = interactions["frac"] > train_size
        if self.session_id_column:
            interactions = self._recalculate_with_session_id_column(interactions)

        train = interactions[~interactions["is_test"]].drop(columns=["row_num", "count", "frac", "is_test"])
        test = interactions[interactions["is_test"]].drop(columns=["row_num", "count", "frac", "is_test"])

        return train, test

    def _partial_split_fractions_spark(
        self, interactions: SparkDataFrame, train_size: float
    ) -> tuple[SparkDataFrame, SparkDataFrame]:
        interactions = interactions.withColumn(
            "count", sf.count(self.timestamp_column).over(Window.partitionBy(self.divide_column))
        )
        if self.min_interactions_per_group is not None:
            interactions = interactions.withColumn(
                "frac",
                sf.when(
                    sf.col("count") >= self.min_interactions_per_group,
                    sf.round(sf.col("row_num") / sf.col("count"), self._precision),
                ).otherwise(sf.lit(0)),
            )
        else:
            interactions = interactions.withColumn(
                "frac", sf.round(sf.col("row_num") / sf.col("count"), self._precision)
            )

        interactions = interactions.withColumn("is_test", sf.col("frac") > train_size)
        if self.session_id_column:
            interactions = self._recalculate_with_session_id_column(interactions)

        train = interactions.filter("is_test == 0").drop("row_num", "count", "frac", "is_test")
        test = interactions.filter("is_test").drop("row_num", "count", "frac", "is_test")

        return train, test

    def _partial_split_fractions_polars(
        self, interactions: PolarsDataFrame, train_size: float
    ) -> tuple[PolarsDataFrame, PolarsDataFrame]:
        interactions = interactions.with_columns(
            pl.count(self.timestamp_column).over(pl.col(self.divide_column)).alias("count")
        )
        if self.min_interactions_per_group is not None:
            interactions = interactions.with_columns(
                pl.when(pl.col("count") >= self.min_interactions_per_group)
                .then((pl.col("row_num") / pl.col("count")).round(self._precision))
                .otherwise(0)
                .alias("frac")
            )
        else:
            interactions = interactions.with_columns(
                (pl.col("row_num") / pl.col("count")).round(self._precision).alias("frac")
            )

        interactions = interactions.with_columns((pl.col("frac") > train_size).alias("is_test"))
        if self.session_id_column:
            interactions = self._recalculate_with_session_id_column(interactions)

        train = interactions.filter(~pl.col("is_test")).drop("row_num", "count", "frac", "is_test")
        test = interactions.filter(pl.col("is_test")).drop("row_num", "count", "frac", "is_test")

        return train, test

    def _partial_split(self, interactions: DataFrameLike, ratio: float) -> tuple[DataFrameLike, DataFrameLike]:
        res = self._add_time_partition(interactions)
        if isinstance(res, SparkDataFrame):
            return self._partial_split_spark(res, ratio)
        if isinstance(res, PandasDataFrame):
            return self._partial_split_pandas(res, ratio)
        else:
            return self._partial_split_polars(res, ratio)

    def _partial_split_pandas(
        self, interactions: PandasDataFrame, ratio: float
    ) -> tuple[PandasDataFrame, PandasDataFrame]:
        interactions["count"] = interactions.groupby(self.divide_column, sort=False)[self.divide_column].transform(len)
        interactions["train_size"] = interactions["count"] - (interactions["count"] * ratio).astype(int)
        if self.min_interactions_per_group is not None:
            interactions["train_size"].where(
                interactions["count"] >= self.min_interactions_per_group, interactions["count"], inplace=True
            )
        else:
            interactions.loc[
                (interactions["count"] * ratio > 0)
                & (interactions["count"] * ratio < 1)
                & (interactions["train_size"] > 1),
                "train_size",
            ] = (
                interactions["train_size"] - 1
            )

        interactions["is_test"] = interactions["row_num"] > interactions["train_size"]
        if self.session_id_column:
            interactions = self._recalculate_with_session_id_column(interactions)

        train = interactions[~interactions["is_test"]].drop(columns=["row_num", "count", "train_size", "is_test"])
        test = interactions[interactions["is_test"]].drop(columns=["row_num", "count", "train_size", "is_test"])

        return train, test

    def _partial_split_spark(self, interactions: SparkDataFrame, ratio: float) -> tuple[SparkDataFrame, SparkDataFrame]:
        interactions = interactions.withColumn(
            "count", sf.count(self.timestamp_column).over(Window.partitionBy(self.divide_column))
        )
        if self.min_interactions_per_group is not None:
            interactions = interactions.withColumn(
                "train_size",
                sf.when(
                    sf.col("count") >= self.min_interactions_per_group,
                    sf.col("count") - (sf.col("count") * ratio).cast(IntegerType()),
                ).otherwise(sf.col("count")),
            )
        else:
            interactions = interactions.withColumn(
                "train_size", sf.col("count") - (sf.col("count") * ratio).cast(IntegerType())
            ).withColumn(
                "train_size",
                sf.when(
                    (sf.col("count") * ratio > 0) & (sf.col("count") * ratio < 1) & (sf.col("train_size") > 1),
                    sf.col("train_size") - 1,
                ).otherwise(sf.col("train_size")),
            )
        interactions = interactions.withColumn("is_test", sf.col("row_num") > sf.col("train_size"))
        if self.session_id_column:
            interactions = self._recalculate_with_session_id_column(interactions)

        train = interactions.filter("is_test == 0").drop("row_num", "count", "train_size", "is_test")
        test = interactions.filter("is_test").drop("row_num", "count", "train_size", "is_test")

        return train, test

    def _partial_split_polars(
        self, interactions: PolarsDataFrame, ratio: float
    ) -> tuple[PolarsDataFrame, PolarsDataFrame]:
        interactions = interactions.with_columns(
            pl.count(self.timestamp_column).over(self.divide_column).alias("count")
        )
        if self.min_interactions_per_group is not None:
            interactions = interactions.with_columns(
                pl.when(pl.col("count") >= self.min_interactions_per_group)
                .then(pl.col("count") - (pl.col("count") * ratio).cast(interactions.get_column("count").dtype))
                .otherwise(pl.col("count"))
                .alias("train_size")
            )
        else:
            interactions = interactions.with_columns(
                (pl.col("count") - (pl.col("count") * ratio).cast(interactions.get_column("count").dtype)).alias(
                    "train_size"
                )
            ).with_columns(
                pl.when((pl.col("count") * ratio > 0) & (pl.col("count") * ratio < 1) & (pl.col("train_size") > 1))
                .then(pl.col("train_size") - 1)
                .otherwise(pl.col("train_size"))
                .alias("train_size")
            )

        interactions = interactions.with_columns((pl.col("row_num") > pl.col("train_size")).alias("is_test"))

        if self.session_id_column:
            interactions = self._recalculate_with_session_id_column(interactions)

        train = interactions.filter(~pl.col("is_test")).drop("row_num", "count", "train_size", "is_test")
        test = interactions.filter(pl.col("is_test")).drop("row_num", "count", "train_size", "is_test")

        return train, test

    def _core_split(self, interactions: DataFrameLike) -> list[DataFrameLike]:
        if self.split_by_fractions:
            return self._partial_split_fractions(interactions, self.test_size)
        else:
            return self._partial_split(interactions, self.test_size)



import math
from typing import Any, Literal, Optional, Union, cast

import lightning
import torch

from replay.data.nn import TensorMap, TensorSchema
from replay.models.nn.loss import ScalableCrossEntropyLoss, SCEParams
from replay.models.nn.optimizer_utils import FatOptimizerFactory, LRSchedulerFactory, OptimizerFactory

from .dataset import SasRecPredictionBatch, SasRecTrainingBatch, SasRecValidationBatch
from .model import SasRecModel


class SasRec(lightning.LightningModule):
    """
    SASRec Lightning module.

    You can get initialization parameters with attribute `hparams`
    for object of SasRec instance.
    """

    def __init__(
        self,
        tensor_schema: TensorSchema,
        block_count: int = 2,
        head_count: int = 1,
        hidden_size: int = 50,
        max_seq_len: int = 200,
        dropout_rate: float = 0.2,
        ti_modification: bool = False,
        time_span: int = 256,
        loss_type: Literal["BCE", "CE", "SCE"] = "CE",
        loss_sample_count: Optional[int] = None,
        negative_sampling_strategy: str = "global_uniform",
        negatives_sharing: bool = False,
        optimizer_factory: OptimizerFactory = FatOptimizerFactory(),
        lr_scheduler_factory: Optional[LRSchedulerFactory] = None,
        sce_params: Optional[SCEParams] = None,
    ):
        """
        :param tensor_schema: Tensor schema of features.
        :param block_count: Number of Transformer blocks.
            Default: ``2``.
        :param head_count: Number of Attention heads.
            Default: ``1``.
        :param hidden_size: Hidden size of transformer.
            Default: ``50``.
        :param max_seq_len: Max length of sequence.
            Default: ``200``.
        :param dropout_rate: Dropout rate.
            Default: ``0.2``.
        :param ti_modification: Enable time relation.
            Default: ``False``.
        :param time_span: Time span value.
            Default: ``256``.
        :param loss_type: Loss type. Possible values: ``"CE"``, ``"BCE"``, ``"SCE"``.
            Default: ``CE``.
        :param loss_sample_count (Optional[int]): Sample count to calculate loss.
            Suitable for ``"CE"`` and ``"BCE"`` loss functions.
            Default: ``None``.
        :param negative_sampling_strategy: Negative sampling strategy to calculate loss on sampled negatives.
            Is used when large count of items in dataset.
            Possible values: ``"global_uniform"``, ``"inbatch"``
            Default: ``global_uniform``.
        :param negatives_sharing: Apply negative sharing in calculating sampled logits.
            Default: ``False``.
        :param optimizer_factory: Optimizer factory.
            Default: ``FatOptimizerFactory``.
        :param lr_scheduler_factory: Learning rate schedule factory.
            Default: ``None``.
        :param sce_params: Dataclass with SCE parameters. Need to be defined if ``loss_type`` is ``SCE``.
            Default: ``None``.
        """
        super().__init__()
        self.save_hyperparameters()
        self._model = SasRecModel(
            schema=tensor_schema,
            num_blocks=block_count,
            num_heads=head_count,
            hidden_size=hidden_size,
            max_len=max_seq_len,
            dropout=dropout_rate,
            ti_modification=ti_modification,
            time_span=time_span,
        )
        self._loss_type = loss_type
        self._loss_sample_count = loss_sample_count
        self._negative_sampling_strategy = negative_sampling_strategy
        self._negatives_sharing = negatives_sharing
        self._optimizer_factory = optimizer_factory
        self._lr_scheduler_factory = lr_scheduler_factory
        self._sce_params = sce_params
        self._loss = self._create_loss()
        self._schema = tensor_schema
        assert negative_sampling_strategy in {"global_uniform", "inbatch"}
        if self._loss_type == "SCE":
            assert sce_params is not None, "You should define ``sce_params`` when using SCE loss function."

        item_count = tensor_schema.item_id_features.item().cardinality
        assert item_count
        self._vocab_size = item_count
        self.candidates_to_score = None

    def training_step(self, batch: SasRecTrainingBatch, batch_idx: int) -> torch.Tensor:
        """
        :param batch (SasRecTrainingBatch): Batch of training data.
        :param batch_idx (int): Batch index.

        :returns: Computed loss for batch.
        """
        if batch_idx % 100 == 0 and torch.cuda.is_available():  # pragma: no cover
            torch.cuda.empty_cache()
        loss = self._compute_loss(batch)
        self.log("train_loss", loss, on_step=True, on_epoch=True, prog_bar=True, sync_dist=True)
        return loss

    def predict_step(
        self,
        batch: SasRecPredictionBatch,
        batch_idx: int,  # noqa: ARG002
        dataloader_idx: int = 0,  # noqa: ARG002
    ) -> torch.Tensor:
        """
        :param batch: Batch of prediction data.
        :param batch_idx: Batch index.
        :param dataloader_idx: Dataloader index.

        :returns: Calculated scores.
        """
        batch = _prepare_prediction_batch(self._schema, self._model.max_len, batch)
        return self._model_predict(batch.features, batch.padding_mask)

    def predict(
        self,
        batch: SasRecPredictionBatch,
        candidates_to_score: Optional[torch.LongTensor] = None,
    ) -> torch.Tensor:
        """
        :param batch: Batch of prediction data.
        :param candidates_to_score: Item ids to calculate scores.
            Default: ``None``.

        :returns: Calculated scores.
        """
        batch = _prepare_prediction_batch(self._schema, self._model.max_len, batch)
        return self._model_predict(batch.features, batch.padding_mask, candidates_to_score)

    def forward(
        self,
        feature_tensors: TensorMap,
        padding_mask: torch.BoolTensor,
        candidates_to_score: Optional[torch.LongTensor] = None,
    ) -> torch.Tensor:  # pragma: no cover
        """
        :param feature_tensors: Batch of features.
        :param padding_mask: Padding mask where 0 - <PAD>, 1 otherwise.
        :param candidates_to_score: Item ids to calculate scores.
            Default: ``None``.

        :returns: Calculated scores.
        """
        return self._model_predict(feature_tensors, padding_mask, candidates_to_score)

    def validation_step(
        self,
        batch: SasRecValidationBatch,
        batch_idx: int,  # noqa: ARG002
        dataloader_idx: int = 0,  # noqa: ARG002
    ) -> torch.Tensor:
        """
        :param batch (SasRecValidationBatch): Batch of prediction data.
        :param batch_idx (int): Batch index.

        :returns: Calculated scores.
        """
        return self._model_predict(batch.features, batch.padding_mask)

    def configure_optimizers(self) -> Any:
        """
        :returns: Configured optimizer and lr scheduler.
        """
        optimizer = self._optimizer_factory.create(self._model.parameters())

        if self._lr_scheduler_factory is None:
            return optimizer

        lr_scheduler = self._lr_scheduler_factory.create(optimizer)
        return [optimizer], [lr_scheduler]

    def _model_predict(
        self,
        feature_tensors: TensorMap,
        padding_mask: torch.BoolTensor,
        candidates_to_score: torch.LongTensor = None,
    ) -> torch.Tensor:
        model: SasRecModel
        model = cast(SasRecModel, self._model.module) if isinstance(self._model, torch.nn.DataParallel) else self._model
        candidates_to_score = self.candidates_to_score if candidates_to_score is None else candidates_to_score
        scores = model.predict(feature_tensors, padding_mask, candidates_to_score)
        return scores

    def _compute_loss(self, batch: SasRecTrainingBatch) -> torch.Tensor:
        if self._loss_type == "BCE":
            loss_func = self._compute_loss_bce if self._loss_sample_count is None else self._compute_loss_bce_sampled
        elif self._loss_type == "CE":
            loss_func = self._compute_loss_ce if self._loss_sample_count is None else self._compute_loss_ce_sampled
        elif self._loss_type == "SCE":
            loss_func = self._compute_loss_scalable_ce
        else:
            msg = f"Not supported loss type: {self._loss_type}"
            raise ValueError(msg)

        loss = loss_func(
            batch.features,
            batch.labels,
            batch.padding_mask,
            batch.labels_padding_mask,
        )
        return loss

    def _compute_loss_bce(
        self,
        feature_tensors: TensorMap,
        positive_labels: torch.LongTensor,
        padding_mask: torch.BoolTensor,
        target_padding_mask: torch.BoolTensor,
    ) -> torch.Tensor:
        # [B x L x V]
        logits = self._model.forward(feature_tensors, padding_mask)

        """
        Take only logits which correspond to non-padded tokens
        M = non_zero_count(target_padding_mask)
        """
        logits = logits[target_padding_mask]  # [M x V]
        labels = positive_labels[target_padding_mask]  # [M]

        bce_labels = torch.zeros(
            (logits.size(0), logits.size(-1)),
            device=logits.device,
        )

        # Fill positives with ones, all negatives are zeros
        bce_labels.scatter_(
            dim=-1,
            index=labels.unsqueeze(-1),
            value=1,
        )

        loss = self._loss(logits, bce_labels) / logits.size(0)
        return loss

    def _compute_loss_bce_sampled(
        self,
        feature_tensors: TensorMap,
        positive_labels: torch.LongTensor,
        padding_mask: torch.BoolTensor,
        target_padding_mask: torch.BoolTensor,
    ) -> torch.Tensor:
        (positive_logits, negative_logits, *_) = self._get_sampled_logits(
            feature_tensors, positive_labels, padding_mask, target_padding_mask
        )

        positive_prob = torch.sigmoid(positive_logits)
        negative_prob = torch.sigmoid(negative_logits)

        # Clamp and eps for numerical stability
        clamp_border: float = 100.0
        eps = 1e-6
        positive_loss = torch.clamp(torch.log((positive_prob) + eps), -clamp_border, clamp_border).sum()
        negative_loss = torch.clamp(torch.log((1 - negative_prob) + eps), -clamp_border, clamp_border).sum()

        loss = -(positive_loss + negative_loss)
        loss /= positive_logits.size(0)

        return loss

    def _compute_loss_ce(
        self,
        feature_tensors: TensorMap,
        positive_labels: torch.LongTensor,
        padding_mask: torch.BoolTensor,
        target_padding_mask: torch.BoolTensor,
    ) -> torch.Tensor:
        # logits: [B x L x V]
        logits = self._model.forward(
            feature_tensors,
            padding_mask,
        )

        # labels: [B x L]
        labels = positive_labels.masked_fill(mask=(~target_padding_mask), value=-100)

        logits_flat = logits.view(-1, logits.size(-1))  # [(B * L) x V]
        labels_flat = labels.view(-1)  # [(B * L)]

        loss = self._loss(logits_flat, labels_flat)
        return loss

    def _compute_loss_ce_sampled(
        self,
        feature_tensors: TensorMap,
        positive_labels: torch.LongTensor,
        padding_mask: torch.BoolTensor,
        target_padding_mask: torch.BoolTensor,
    ) -> torch.Tensor:
        assert self._loss_sample_count is not None
        (positive_logits, negative_logits, positive_labels, negative_labels, vocab_size) = self._get_sampled_logits(
            feature_tensors, positive_labels, padding_mask, target_padding_mask
        )
        n_negative_samples = min(self._loss_sample_count, vocab_size)

        # Reject negative samples matching target label & correct for remaining samples
        reject_labels = positive_labels == negative_labels  # [masked_batch_seq_size x n_negative_samples]
        negative_logits += math.log(vocab_size - 1)
        negative_logits -= 1e6 * reject_labels
        negative_logits -= torch.log((n_negative_samples - reject_labels.sum(dim=-1, keepdim=True)).float())

        # Apply regular softmax cross entropy
        # [masked_batch_seq_size x (1 + n_negative_samples)]
        logits = torch.cat([positive_logits, negative_logits], dim=1).float()
        labels_flat = torch.zeros(positive_logits.size(0), dtype=torch.long, device=padding_mask.device)
        loss = self._loss(logits, labels_flat)
        return loss

    def _compute_loss_scalable_ce(
        self,
        feature_tensors: TensorMap,
        positive_labels: torch.LongTensor,
        padding_mask: torch.BoolTensor,
        tokens_mask: torch.BoolTensor,  # noqa: ARG002
    ) -> torch.Tensor:
        emb = self._model.forward_step(feature_tensors, padding_mask)
        all_embeddings = self.get_all_embeddings()["item_embedding"]
        return self._loss(emb, positive_labels, all_embeddings, padding_mask)

    def _get_sampled_logits(
        self,
        feature_tensors: TensorMap,
        positive_labels: torch.LongTensor,
        padding_mask: torch.BoolTensor,
        target_padding_mask: torch.BoolTensor,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.LongTensor, torch.LongTensor, int]:
        assert self._loss_sample_count is not None
        n_negative_samples = self._loss_sample_count
        positive_labels = cast(
            torch.LongTensor, torch.masked_select(positive_labels, target_padding_mask)
        )  # (masked_batch_seq_size,)
        masked_batch_seq_size = positive_labels.size(0)
        device = padding_mask.device
        output_emb = self._model.forward_step(feature_tensors, padding_mask)[target_padding_mask]

        positive_labels = cast(torch.LongTensor, positive_labels.view(-1, 1))
        ids = torch.arange(masked_batch_seq_size, dtype=torch.long, device=device)
        unique_positive_labels, positive_labels_indices = positive_labels.unique(return_inverse=True)

        if self._negative_sampling_strategy == "global_uniform":
            vocab_size = self._vocab_size
            multinomial_sample_distribution = torch.ones(vocab_size, device=device)
            # positive_labels - 2d
            positive_logits = self._model.get_logits(output_emb, positive_labels)
        elif self._negative_sampling_strategy == "inbatch":
            positive_labels_indices = positive_labels_indices.view(masked_batch_seq_size, 1)
            # unique_positive_labels - 1d
            positive_logits = self._model.get_logits(output_emb, unique_positive_labels)
            vocab_size = unique_positive_labels.size(0)
            if self._negatives_sharing:
                multinomial_sample_distribution = torch.ones(vocab_size, device=device)
            else:
                multinomial_sample_distribution = torch.softmax(positive_logits, dim=-1)
        else:
            msg = f"Unknown negative sampling strategy: {self._negative_sampling_strategy}"
            raise NotImplementedError(msg)
        n_negative_samples = min(n_negative_samples, vocab_size)

        if self._negatives_sharing:
            negative_labels = torch.multinomial(
                multinomial_sample_distribution,
                num_samples=n_negative_samples,
                replacement=False,
            )
            negative_labels = negative_labels.unsqueeze(0).repeat(masked_batch_seq_size, 1)
        elif self._negative_sampling_strategy == "global_uniform":
            negative_labels = torch.randint(
                low=0,
                high=vocab_size,
                size=(masked_batch_seq_size, n_negative_samples),
                dtype=torch.long,
                device=device,
            )
        else:
            negative_labels = torch.multinomial(
                multinomial_sample_distribution,
                num_samples=n_negative_samples,
                replacement=False,
            )
        negative_labels = cast(torch.LongTensor, negative_labels)

        if self._negative_sampling_strategy == "global_uniform":
            if self._negatives_sharing:
                unique_negative_labels, negative_labels_indices = negative_labels.unique(return_inverse=True)
                negative_labels_indices = negative_labels_indices.view(masked_batch_seq_size, n_negative_samples)
                # unique_negative_labels - 1d
                negative_logits = self._model.get_logits(output_emb, unique_negative_labels)
                negative_logits = negative_logits[ids, negative_labels_indices.T].T
            else:
                # unique_negative_labels - 1d
                negative_logits = self._model.get_logits(output_emb, negative_labels)
        else:  # self._negative_sampling_strategy == "inbatch":
            negative_labels_indices = negative_labels
            negative_logits = positive_logits
            negative_logits = negative_logits[ids, negative_labels_indices.T].T
            positive_logits = positive_logits[ids, positive_labels_indices.T].T

        return (positive_logits, negative_logits, positive_labels, negative_labels, vocab_size)

    def _create_loss(self) -> Union[torch.nn.BCEWithLogitsLoss, torch.nn.CrossEntropyLoss]:
        if self._loss_type == "BCE":
            return torch.nn.BCEWithLogitsLoss(reduction="sum")

        if self._loss_type == "CE":
            return torch.nn.CrossEntropyLoss()

        if self._loss_type == "SCE":
            return ScalableCrossEntropyLoss(self._sce_params)

        msg = "Not supported loss_type"
        raise NotImplementedError(msg)

    def get_all_embeddings(self) -> dict[str, torch.nn.Embedding]:
        """
        :returns: copy of all embeddings as a dictionary.
        """
        return self._model.item_embedder.get_all_embeddings()

    def set_item_embeddings_by_size(self, new_vocab_size: int):
        """
        Keep the current item embeddings and expand vocabulary with new embeddings
        initialized with xavier_normal_ for new items.

        :param new_vocab_size: Size of vocabulary with new items included.
            Must be greater then already fitted.
        """
        old_vocab_size = self._model.item_embedder.item_emb.weight.data.shape[0] - 1
        hidden_size = self._model.hidden_size

        if new_vocab_size <= old_vocab_size:
            msg = "New vocabulary size must be greater then already fitted"
            raise ValueError(msg)

        new_embedding = torch.nn.Embedding(new_vocab_size + 1, hidden_size, padding_idx=new_vocab_size)
        torch.nn.init.xavier_normal_(new_embedding.weight)
        new_embedding.weight.data[:old_vocab_size, :] = self._model.item_embedder.item_emb.weight.data[:-1, :]

        self._set_new_item_embedder_to_model(new_embedding, new_vocab_size)

    def set_item_embeddings_by_tensor(self, all_item_embeddings: torch.Tensor):
        """
        Set item embeddings with provided weights for all items.
        If new items presented, then tensor is expanded.
        The already fitted weights will be replaced with new ones.

        :param all_item_embeddings: tensor of weights for all items with
            shape (n, h), where n - number of all items, h - model hidden size.
        """
        if all_item_embeddings.dim() != 2:
            msg = "Input tensor must have (number of all items, model hidden size) shape"
            raise ValueError(msg)

        old_vocab_size = self._model.item_embedder.item_emb.weight.data.shape[0] - 1
        new_vocab_size = all_item_embeddings.shape[0]
        hidden_size = self._model.hidden_size

        if new_vocab_size < old_vocab_size:
            msg = "New vocabulary size can't be less then already fitted"
            raise ValueError(msg)
        if all_item_embeddings.shape[1] != hidden_size:
            msg = "Input tensor second dimension doesn't match model hidden size"
            raise ValueError(msg)

        new_embedding = torch.nn.Embedding(new_vocab_size + 1, hidden_size, padding_idx=new_vocab_size)
        new_embedding.weight.data[:-1, :] = all_item_embeddings

        self._set_new_item_embedder_to_model(new_embedding, new_vocab_size)

    def append_item_embeddings(self, item_embeddings: torch.Tensor):
        """
        Append provided weights for new items only to item embedder.

        :param item_embeddings: tensor of shape (n, h), where
            n - number of only new items, h - model hidden size.
        """
        if item_embeddings.dim() != 2:
            msg = "Input tensor must have (number of new items, model hidden size) shape"
            raise ValueError(msg)

        old_vocab_size = self._model.item_embedder.item_emb.weight.data.shape[0] - 1
        new_vocab_size = item_embeddings.shape[0] + old_vocab_size
        hidden_size = self._model.hidden_size

        if item_embeddings.shape[1] != hidden_size:
            msg = "Input tensor second dimension doesn't match model hidden size"
            raise ValueError(msg)

        new_embedding = torch.nn.Embedding(new_vocab_size + 1, hidden_size, padding_idx=new_vocab_size)
        new_embedding.weight.data[:old_vocab_size, :] = self._model.item_embedder.item_emb.weight.data[:-1, :]
        new_embedding.weight.data[old_vocab_size:-1, :] = item_embeddings

        self._set_new_item_embedder_to_model(new_embedding, new_vocab_size)

    @property
    def optimizer_factory(self) -> OptimizerFactory:
        """
        Returns current optimizer_factory.
        """
        return self._optimizer_factory

    @optimizer_factory.setter
    def optimizer_factory(self, optimizer_factory: OptimizerFactory) -> None:
        """
        Sets new optimizer_factory.
        :param optimizer_factory: New optimizer factory.
        """
        if isinstance(optimizer_factory, OptimizerFactory):
            self._optimizer_factory = optimizer_factory
        else:
            msg = f"Expected optimizer_factory of type OptimizerFactory, got {type(optimizer_factory)}"
            raise ValueError(msg)

    @property
    def candidates_to_score(self) -> Union[torch.LongTensor, None]:
        """
        Returns tensor of item ids to calculate scores.
        """
        return self._candidates_to_score

    @candidates_to_score.setter
    def candidates_to_score(self, candidates: Optional[torch.LongTensor] = None) -> None:
        """
        Sets tensor of item ids to calculate scores.
        :param candidates: Tensor of item ids to calculate scores.
        """
        total_item_count = self._model.item_count
        if isinstance(candidates, torch.Tensor) and candidates.dtype is torch.long:
            if 0 < candidates.shape[0] <= total_item_count:
                self._candidates_to_score = candidates
            else:
                msg = f"Expected candidates length to be between 1 and {total_item_count=}"
                raise ValueError(msg)
        elif candidates is not None:
            msg = f"Expected candidates to be of type torch.LongTensor or None, gpt {type(candidates)}"
            raise ValueError(msg)
        self._candidates_to_score = candidates

    def _set_new_item_embedder_to_model(self, new_embedding: torch.nn.Embedding, new_vocab_size: int):
        self._model.item_embedder.item_emb = new_embedding
        self._model._head._item_embedder = self._model.item_embedder
        self._vocab_size = new_vocab_size
        self._model.item_count = new_vocab_size
        self._model.padding_idx = new_vocab_size
        self._model.masking.padding_idx = new_vocab_size
        self._schema.item_id_features[self._schema.item_id_feature_name]._set_cardinality(
            new_embedding.weight.data.shape[0] - 1
        )


def _prepare_prediction_batch(
    schema: TensorSchema, max_len: int, batch: SasRecPredictionBatch
) -> SasRecPredictionBatch:
    if batch.padding_mask.shape[1] > max_len:
        msg = (
            "The length of the submitted sequence "
            "must not exceed the maximum length of the sequence. "
            f"The length of the sequence is given {batch.padding_mask.shape[1]}, "
            f"while the maximum length is {max_len}"
        )
        raise ValueError(msg)

    if batch.padding_mask.shape[1] < max_len:
        query_id, padding_mask, features = batch
        sequence_item_count = padding_mask.shape[1]
        for feature_name, feature_tensor in features.items():
            if schema[feature_name].is_cat:
                features[feature_name] = torch.nn.functional.pad(
                    feature_tensor, (max_len - sequence_item_count, 0), value=0
                )
            else:
                features[feature_name] = torch.nn.functional.pad(
                    feature_tensor.view(feature_tensor.size(0), feature_tensor.size(1)),
                    (max_len - sequence_item_count, 0),
                    value=0,
                ).unsqueeze(-1)
        padding_mask = torch.nn.functional.pad(padding_mask, (max_len - sequence_item_count, 0), value=0)
        batch = SasRecPredictionBatch(query_id, padding_mask, features)
    return batch



from collections.abc import Iterable
from typing import Optional

import polars as pl

from replay.utils import PYSPARK_AVAILABLE, DataFrameLike, PandasDataFrame, PolarsDataFrame
from replay.utils.spark_utils import spark_to_pandas

if PYSPARK_AVAILABLE:  # pragma: no cover
    import pyspark.sql.functions as sf


def groupby_sequences(events: DataFrameLike, groupby_col: str, sort_col: Optional[str] = None) -> DataFrameLike:
    """
    :param events: dataframe with interactions
    :param groupby_col: divide column to group by
    :param sort_col: column to sort by

    :returns: dataframe with sequences for each value in groupby_col
    """
    if isinstance(events, PandasDataFrame):
        event_cols_without_groupby = events.columns.values.tolist()
        event_cols_without_groupby.remove(groupby_col)

        if sort_col:
            event_cols_without_iterable = list(
                filter(lambda x: not isinstance(events.iloc[0][x], Iterable), event_cols_without_groupby)
            )  # deleting columns that cannot be sorted
            event_cols_without_iterable.remove(sort_col)
            event_cols_without_iterable.insert(0, sort_col)
            events = events.sort_values(event_cols_without_iterable)

        grouped_sequences = (
            events.groupby(groupby_col).agg(dict.fromkeys(event_cols_without_groupby, list)).reset_index()
        )
    elif isinstance(events, PolarsDataFrame):
        event_cols_without_groupby = events.columns
        event_cols_without_groupby.remove(groupby_col)

        if sort_col:
            map_name2type = dict(zip(events.columns, events.dtypes))
            event_cols_without_iterable = list(
                filter(lambda x: not isinstance(map_name2type[x], pl.List), event_cols_without_groupby)
            )  # deleting columns that cannot be sorted
            event_cols_without_iterable.remove(sort_col)
            event_cols_without_iterable.insert(0, sort_col)
            events = events.sort(event_cols_without_iterable)

        grouped_sequences = events.group_by(groupby_col).agg(*[pl.col(x) for x in event_cols_without_groupby])
    else:
        event_cols_without_groupby = events.columns.copy()
        event_cols_without_groupby.remove(groupby_col)

        if sort_col:
            event_cols_without_groupby.remove(sort_col)
            event_cols_without_groupby.insert(0, sort_col)

        all_cols_struct = sf.struct(event_cols_without_groupby)

        collect_fn = sf.collect_list(all_cols_struct)
        if sort_col:
            collect_fn = sf.sort_array(collect_fn)

        grouped_sequences = (
            events.groupby(groupby_col)
            .agg(collect_fn.alias("_"))
            .select([sf.col(groupby_col)] + [sf.col(f"_.{col}").alias(col) for col in event_cols_without_groupby])
            .drop("_")
        )

    return grouped_sequences


def ensure_pandas(
    df: DataFrameLike,
    allow_collect_to_master: bool = False,
) -> PandasDataFrame:
    """
    :param df: dataframe
    :param allow_collect_to_master: Flag allowing spark to make a collection to the master node,
        default: ``False``.

    :returns: Pandas DataFrame object
    """
    if isinstance(df, PandasDataFrame):
        return df
    return spark_to_pandas(df, allow_collect_to_master)



import warnings
from collections.abc import Sequence
from copy import deepcopy
from functools import partial
from typing import NoReturn, Optional, Union

from typing_extensions import TypeAlias

from replay.data import Dataset
from replay.metrics import NDCG, Metric
from replay.models.common import RecommenderCommons
from replay.models.optimization.optuna_objective import ObjectiveWrapper, SplitData, scenario_objective_calculator
from replay.utils import OPTUNA_AVAILABLE, FeatureUnavailableError, FeatureUnavailableWarning

MainObjective = partial(ObjectiveWrapper, objective_calculator=scenario_objective_calculator)

if OPTUNA_AVAILABLE:

    class OptunaMixin(RecommenderCommons):
        """
        A mixin class enabling hyperparameter optimization in a recommender using Optuna objectives.
        """

        _objective = MainObjective
        _search_space: Optional[dict[str, Union[str, Sequence[Union[str, int, float]]]]] = None
        study = None
        criterion: Optional[Metric] = None

        @staticmethod
        def _filter_dataset_features(
            dataset: Dataset,
        ) -> Dataset:
            """
            Filter features of dataset to match with items and queries of interactions

            :param dataset: dataset with interactions and features
            :return: filtered dataset
            """
            if dataset.query_features is None and dataset.item_features is None:
                return dataset

            query_features = None
            item_features = None
            if dataset.query_features is not None:
                query_features = dataset.query_features.join(
                    dataset.interactions.select(dataset.feature_schema.query_id_column).distinct(),
                    on=dataset.feature_schema.query_id_column,
                )
            if dataset.item_features is not None:
                item_features = dataset.item_features.join(
                    dataset.interactions.select(dataset.feature_schema.item_id_column).distinct(),
                    on=dataset.feature_schema.item_id_column,
                )

            return Dataset(
                feature_schema=dataset.feature_schema,
                interactions=dataset.interactions,
                query_features=query_features,
                item_features=item_features,
                check_consistency=False,
                categorical_encoded=False,
            )

        def _prepare_split_data(
            self,
            train_dataset: Dataset,
            test_dataset: Dataset,
        ) -> SplitData:
            """
            This method converts data to spark and packs it into a named tuple to pass into optuna.

            :param train_dataset: train data
            :param test_dataset: test data
            :return: packed PySpark DataFrames
            """
            train = self._filter_dataset_features(train_dataset)
            test = self._filter_dataset_features(test_dataset)
            queries = test_dataset.interactions.select(self.query_column).distinct()
            items = test_dataset.interactions.select(self.item_column).distinct()

            split_data = SplitData(
                train,
                test,
                queries,
                items,
            )
            return split_data

        def _check_borders(self, param, borders):
            """Raise value error if param borders are not valid"""
            if param not in self._search_space:
                msg = f"Hyper parameter {param} is not defined for {self!s}"
                raise ValueError(msg)
            if not isinstance(borders, list):
                msg = f"Parameter {param} borders are not a list"
                raise ValueError()
            if self._search_space[param]["type"] != "categorical" and len(borders) != 2:
                msg = f"Hyper parameter {param} is numerical but bounds are not in ([lower, upper]) format"
                raise ValueError(msg)

        def _prepare_param_borders(self, param_borders: Optional[dict[str, list]] = None) -> dict[str, dict[str, list]]:
            """
            Checks if param borders are valid and convert them to a search_space format

            :param param_borders: a dictionary with search grid, where
                key is the parameter name and value is the range of possible values
                ``{param: [low, high]}``.
            :return:
            """
            search_space = deepcopy(self._search_space)
            if param_borders is None:
                return search_space

            for param, borders in param_borders.items():
                self._check_borders(param, borders)
                search_space[param]["args"] = borders

            # Optuna trials should contain all searchable parameters
            # to be able to correctly return best params
            # If used didn't specify some params to be tested optuna still needs to suggest them
            # This part makes sure this suggestion will be constant
            args = self._init_args
            missing_borders = {param: args[param] for param in search_space if param not in param_borders}
            for param, value in missing_borders.items():
                if search_space[param]["type"] == "categorical":
                    search_space[param]["args"] = [value]
                else:
                    search_space[param]["args"] = [value, value]

            return search_space

        def _init_params_in_search_space(self, search_space):
            """Check if model params are inside search space"""
            params = self._init_args
            outside_search_space = {}
            for param, value in params.items():
                if param not in search_space:
                    continue
                borders = search_space[param]["args"]
                param_type = search_space[param]["type"]

                extra_category = param_type == "categorical" and value not in borders
                param_out_of_bounds = param_type != "categorical" and (value < borders[0] or value > borders[1])
                if extra_category or param_out_of_bounds:
                    outside_search_space[param] = {
                        "borders": borders,
                        "value": value,
                    }

            if outside_search_space:
                self.logger.debug(
                    "Model is initialized with parameters outside the search space: %s."
                    "Initial parameters will not be evaluated during optimization."
                    "Change search spare with 'param_borders' argument if necessary",
                    outside_search_space,
                )
                return False
            else:
                return True

        def _params_tried(self):
            """check if current parameters were already evaluated"""
            if self.study is None:
                return False

            params = {name: value for name, value in self._init_args.items() if name in self._search_space}
            return any(params == trial.params for trial in self.study.trials)

        def optimize(
            self,
            train_dataset: Dataset,
            test_dataset: Dataset,
            param_borders: Optional[dict[str, list]] = None,
            criterion: Metric = NDCG,
            k: int = 10,
            budget: int = 10,
            new_study: bool = True,
        ) -> Optional[dict]:
            """
            Searches the best parameters with optuna.

            :param train_dataset: train data
            :param test_dataset: test data
            :param param_borders: a dictionary with search borders, where
                key is the parameter name and value is the range of possible values
                ``{param: [low, high]}``. In case of categorical parameters it is
                all possible values: ``{cat_param: [cat_1, cat_2, cat_3]}``.
            :param criterion: metric to use for optimization
            :param k: recommendation list length
            :param budget: number of points to try
            :param new_study: keep searching with previous study or start a new study
            :return: dictionary with best parameters
            """
            from optuna import create_study
            from optuna.samplers import TPESampler

            self.query_column = train_dataset.feature_schema.query_id_column
            self.item_column = train_dataset.feature_schema.item_id_column
            self.rating_column = train_dataset.feature_schema.interactions_rating_column
            self.timestamp_column = train_dataset.feature_schema.interactions_timestamp_column

            self.criterion = criterion(
                topk=k,
                query_column=self.query_column,
                item_column=self.item_column,
                rating_column=self.rating_column,
            )

            if self._search_space is None:
                self.logger.warning("%s has no hyper parameters to optimize", str(self))
                return None

            if self.study is None or new_study:
                self.study = create_study(direction="maximize", sampler=TPESampler())

            search_space = self._prepare_param_borders(param_borders)
            if self._init_params_in_search_space(search_space) and not self._params_tried():
                self.study.enqueue_trial(self._init_args)

            split_data = self._prepare_split_data(train_dataset, test_dataset)
            objective = self._objective(
                search_space=search_space,
                split_data=split_data,
                recommender=self,
                criterion=self.criterion,
                k=k,
            )

            self.study.optimize(objective, budget)
            best_params = self.study.best_params
            self.set_params(**best_params)
            return best_params

else:
    feature_warning = FeatureUnavailableWarning(
        "Optimization feature not enabled - `optuna` package not found. "
        "Ensure you have the package installed if you want to "
        "use the `optimize()` method in your recommenders."
    )
    warnings.warn(feature_warning)

    class OptunaStub(RecommenderCommons):
        """A stub class to use in case of missing dependencies."""

        def optimize(
            self,
            train_dataset: Dataset,  # noqa: ARG002
            test_dataset: Dataset,  # noqa: ARG002
            param_borders: Optional[dict[str, list]] = None,  # noqa: ARG002
            criterion: Metric = NDCG,  # noqa: ARG002
            k: int = 10,  # noqa: ARG002
            budget: int = 10,  # noqa: ARG002
            new_study: bool = True,  # noqa: ARG002
        ) -> NoReturn:
            """
            Searches the best parameters with optuna.

            :param train_dataset: train data
            :param test_dataset: test data
            :param param_borders: a dictionary with search borders, where
                key is the parameter name and value is the range of possible values
                ``{param: [low, high]}``. In case of categorical parameters it is
                all possible values: ``{cat_param: [cat_1, cat_2, cat_3]}``.
            :param criterion: metric to use for optimization
            :param k: recommendation list length
            :param budget: number of points to try
            :param new_study: keep searching with previous study or start a new study
            :return: dictionary with best parameters
            """
            import sys

            err = FeatureUnavailableError('Cannot use method "optimize()" - Optuna not found.')
            if sys.version_info >= (3, 11):  # pragma: py-lt-311
                err.add_note('To enable this functionality, ensure you have the "optuna" package isntalled.')

            raise err


IsOptimizible: TypeAlias = OptunaMixin if OPTUNA_AVAILABLE else OptunaStub



"""Distribution calculations"""

from .types import PYSPARK_AVAILABLE, DataFrameLike, PandasDataFrame

if PYSPARK_AVAILABLE:
    from pyspark.sql import functions as sf

    from replay.utils.spark_utils import convert2spark, get_top_k_recs, spark_to_pandas


def item_distribution(
    log: DataFrameLike,
    recommendations: DataFrameLike,
    k: int,
    allow_collect_to_master: bool = False,
) -> PandasDataFrame:
    """
    Calculate item distribution in ``log`` and ``recommendations``.

    :param log: historical DataFrame used to calculate popularity
    :param recommendations: model recommendations
    :param k: length of a recommendation list
    :return: DataFrame with results
    """
    log = convert2spark(log)
    res = log.groupBy("item_idx").agg(sf.countDistinct("user_idx").alias("user_count")).select("item_idx", "user_count")

    rec = convert2spark(recommendations)
    rec = get_top_k_recs(rec, k)
    rec = rec.groupBy("item_idx").agg(sf.countDistinct("user_idx").alias("rec_count")).select("item_idx", "rec_count")

    res = res.join(rec, on="item_idx", how="outer").fillna(0).orderBy(["user_count", "item_idx"])
    return spark_to_pandas(res, allow_collect_to_master)



from typing import Optional

import polars as pl

from replay.utils import PYSPARK_AVAILABLE, DataFrameLike, PandasDataFrame, PolarsDataFrame, SparkDataFrame

from .base_splitter import Splitter, SplitterReturnType

if PYSPARK_AVAILABLE:
    import pyspark.sql.functions as sf
    from pyspark.sql import Window


class NewUsersSplitter(Splitter):
    """
    Only new users will be assigned to test set.
    Splits interactions by timestamp so that test has `test_size` fraction of most recent users.


    >>> from replay.splitters import NewUsersSplitter
    >>> import pandas as pd
    >>> data_frame = pd.DataFrame({"query_id": [1,1,2,2,3,4],
    ...    "item_id": [1,2,3,1,2,3],
    ...    "relevance": [1,2,3,4,5,6],
    ...    "timestamp": [20,40,20,30,10,40]})
    >>> data_frame
       query_id   item_id  relevance  timestamp
    0         1         1          1         20
    1         1         2          2         40
    2         2         3          3         20
    3         2         1          4         30
    4         3         2          5         10
    5         4         3          6         40
    >>> train, test = NewUsersSplitter(test_size=0.1).split(data_frame)
    >>> train
      query_id  item_id  relevance  timestamp
    0        1        1          1         20
    2        2        3          3         20
    3        2        1          4         30
    4        3        2          5         10
    <BLANKLINE>
    >>> test
      query_id  item_id  relevance  timestamp
    0        4        3          6         40
    <BLANKLINE>

    Train DataFrame can be drastically reduced even with moderate
    `test_size` if the amount of new users is small.

    >>> train, test = NewUsersSplitter(test_size=0.3).split(data_frame)
    >>> train
      query_id  item_id  relevance  timestamp
    4        3        2          5         10
    <BLANKLINE>
    """

    _init_arg_names = [
        "test_size",
        "drop_cold_items",
        "query_column",
        "item_column",
        "timestamp_column",
        "session_id_column",
        "session_id_processing_strategy",
    ]

    def __init__(
        self,
        test_size: float,
        drop_cold_items: bool = False,
        query_column: str = "query_id",
        item_column: Optional[str] = "item_id",
        timestamp_column: Optional[str] = "timestamp",
        session_id_column: Optional[str] = None,
        session_id_processing_strategy: str = "test",
    ):
        """
        :param test_size: test size 0 to 1
        :param drop_cold_items: flag to drop cold items from test
        :param query_column: query id column name
        :param item_column: item id column name
        :param timestamp_column: timestamp column name
        :param session_id_column: name of session id column, which values can not be split.
        :param session_id_processing_strategy: strategy of processing session if it is split,
            values: ``train, test``, train: whole split session goes to train. test: same but to test.
            default: ``test``.
        """
        super().__init__(
            drop_cold_items=drop_cold_items,
            query_column=query_column,
            item_column=item_column,
            timestamp_column=timestamp_column,
            session_id_column=session_id_column,
            session_id_processing_strategy=session_id_processing_strategy,
        )
        if test_size < 0 or test_size > 1:
            msg = "test_size must between 0 and 1"
            raise ValueError(msg)
        self.test_size = test_size

    def _core_split_pandas(
        self, interactions: PandasDataFrame, threshold: float
    ) -> tuple[PandasDataFrame, PandasDataFrame]:
        start_date_by_user = (
            interactions.groupby(self.query_column).agg(_start_dt_by_user=(self.timestamp_column, "min")).reset_index()
        )
        test_start_date = (
            start_date_by_user.groupby("_start_dt_by_user")
            .agg(_num_users_by_start_date=(self.query_column, "count"))
            .reset_index()
            .sort_values(by="_start_dt_by_user", ascending=False)
        )
        test_start_date["_cum_num_users_to_dt"] = test_start_date["_num_users_by_start_date"].cumsum()
        test_start_date["total"] = sum(test_start_date["_num_users_by_start_date"])
        test_start_date = test_start_date[
            test_start_date["_cum_num_users_to_dt"] >= threshold * test_start_date["total"]
        ]
        test_start = test_start_date["_start_dt_by_user"].max()

        train = interactions[interactions[self.timestamp_column] < test_start]
        test = interactions.merge(
            start_date_by_user[start_date_by_user["_start_dt_by_user"] >= test_start], how="inner", on=self.query_column
        ).drop(columns=["_start_dt_by_user"])

        if self.session_id_column:
            interactions["is_test"] = False
            interactions.loc[test.index, "is_test"] = True
            interactions = self._recalculate_with_session_id_column(interactions)
            train = interactions[~interactions["is_test"]].drop(columns=["is_test"])
            test = interactions[interactions["is_test"]].drop(columns=["is_test"])
            interactions = interactions.drop(columns=["is_test"])

        return train, test

    def _core_split_spark(
        self, interactions: SparkDataFrame, threshold: float
    ) -> tuple[SparkDataFrame, SparkDataFrame]:
        start_date_by_user = interactions.groupby(self.query_column).agg(
            sf.min(self.timestamp_column).alias("_start_dt_by_user")
        )
        test_start_date = (
            start_date_by_user.groupby("_start_dt_by_user")
            .agg(sf.count(self.query_column).alias("_num_users_by_start_date"))
            .select(
                "_start_dt_by_user",
                sf.sum("_num_users_by_start_date")
                .over(Window.orderBy(sf.desc("_start_dt_by_user")))
                .alias("_cum_num_users_to_dt"),
                sf.sum("_num_users_by_start_date").over(Window.orderBy(sf.lit(1))).alias("total"),
            )
            .filter(sf.col("_cum_num_users_to_dt") >= sf.col("total") * threshold)
            .agg(sf.max("_start_dt_by_user"))
            .head()[0]
        )

        train = interactions.filter(sf.col(self.timestamp_column) < test_start_date)
        test = interactions.join(
            start_date_by_user.filter(sf.col("_start_dt_by_user") >= test_start_date),
            how="inner",
            on=self.query_column,
        ).drop("_start_dt_by_user")

        if self.session_id_column:
            test = test.withColumn("is_test", sf.lit(True))
            interactions = interactions.join(test, on=interactions.schema.names, how="left").na.fill({"is_test": False})
            interactions = self._recalculate_with_session_id_column(interactions)
            train = interactions.filter(~sf.col("is_test")).drop("is_test")
            test = interactions.filter(sf.col("is_test")).drop("is_test")

        return train, test

    def _core_split_polars(
        self, interactions: PolarsDataFrame, threshold: float
    ) -> tuple[PolarsDataFrame, PolarsDataFrame]:
        start_date_by_user = interactions.group_by(self.query_column).agg(
            pl.col(self.timestamp_column).min().alias("_start_dt_by_user")
        )
        test_start_date = (
            start_date_by_user.group_by("_start_dt_by_user")
            .agg(pl.col(self.query_column).count().alias("_num_users_by_start_date"))
            .sort("_start_dt_by_user", descending=True)
            .with_columns(
                pl.col("_num_users_by_start_date").cum_sum().alias("cum_sum_users"),
            )
            .filter(pl.col("cum_sum_users") >= pl.col("cum_sum_users").max() * threshold)["_start_dt_by_user"]
            .max()
        )

        train = interactions.filter(pl.col(self.timestamp_column) < test_start_date)
        test = interactions.join(
            start_date_by_user.filter(pl.col("_start_dt_by_user") >= test_start_date), on=self.query_column, how="inner"
        ).drop("_start_dt_by_user")

        if self.session_id_column:
            interactions = interactions.with_columns(
                pl.when(pl.col(self.timestamp_column) < test_start_date).then(False).otherwise(True).alias("is_test")
            )
            interactions = self._recalculate_with_session_id_column(interactions)
            train = interactions.filter(~pl.col("is_test")).drop("is_test")
            test = interactions.filter(pl.col("is_test")).drop("is_test")

        return train, test

    def _core_split(self, interactions: DataFrameLike) -> SplitterReturnType:
        if isinstance(interactions, SparkDataFrame):
            return self._core_split_spark(interactions, self.test_size)
        if isinstance(interactions, PandasDataFrame):
            return self._core_split_pandas(interactions, self.test_size)
        if isinstance(interactions, PolarsDataFrame):
            return self._core_split_polars(interactions, self.test_size)

        msg = f"{self} is not implemented for {type(interactions)}"
        raise NotImplementedError(msg)



from replay.utils import OPENVINO_AVAILABLE

if OPENVINO_AVAILABLE:
    from .bert4rec_compiled import Bert4RecCompiled
    from .sasrec_compiled import SasRecCompiled

    __all__ = ["Bert4RecCompiled", "SasRecCompiled"]
else:
    import sys

    err = ImportError('Cannot import from module "compiled" - OpenVINO prerequisites not found.')
    if sys.version_info >= (3, 11):  # pragma: py-lt-311
        err.add_note('To enable this functionality, ensure you have both "openvino" and "onnx" packages isntalled.')

    raise err



import pathlib
from typing import Optional, Union, get_args

import openvino as ov
import torch

from replay.data.nn import TensorSchema
from replay.models.nn.sequential.compiled.base_compiled_model import (
    BaseCompiledModel,
    OptimizedModeType,
)
from replay.models.nn.sequential.sasrec import (
    SasRec,
    SasRecPredictionBatch,
)
from replay.models.nn.sequential.sasrec.lightning import _prepare_prediction_batch


class SasRecCompiled(BaseCompiledModel):
    """
    SasRec CPU-optimized model for inference via OpenVINO.
    It is recommended to compile model with ``compile`` method and pass ``SasRec`` checkpoint
    or the model object itself into it.
    It is also possible to compile model by yourself and pass it to the ``__init__`` with ``TensorSchema``.

    **Note** that compilation requires disk write (and maybe delete) permission.
    """

    def __init__(
        self,
        compiled_model: ov.CompiledModel,
        schema: TensorSchema,
    ) -> None:
        """
        :param compiled_model: Compiled model.
        :param schema: Tensor schema of SasRec model.
        """
        super().__init__(compiled_model, schema)

    def predict(
        self,
        batch: SasRecPredictionBatch,
        candidates_to_score: Optional[torch.LongTensor] = None,
    ) -> torch.Tensor:
        """
        Inference on one batch.

        :param batch: Prediction input.
        :param candidates_to_score: Item ids to calculate scores.
            Default: ``None``.

        :return: Tensor with scores.
        """
        self._valilade_predict_input(batch, candidates_to_score)

        batch = _prepare_prediction_batch(self._schema, self._max_seq_len, batch)
        model_inputs = {
            self._inputs_names[0]: batch.features[self._inputs_names[0]],
            self._inputs_names[1]: batch.padding_mask,
        }
        if self._num_candidates_to_score is not None:
            self._validate_candidates_to_score(candidates_to_score)
            model_inputs[self._inputs_names[2]] = candidates_to_score
        return torch.from_numpy(self._model(model_inputs)[self._output_name])

    @classmethod
    def compile(
        cls,
        model: Union[SasRec, str, pathlib.Path],
        mode: OptimizedModeType = "one_query",
        batch_size: Optional[int] = None,
        num_candidates_to_score: Optional[int] = None,
        num_threads: Optional[int] = None,
        onnx_path: Optional[str] = None,
    ) -> "SasRecCompiled":
        """
        Model compilation.

        :param model: Path to lightning SasRec model saved in .ckpt format or the SasRec object itself.
        :param mode: Inference mode, defines shape of inputs.
            Could be one of [``one_query``, ``batch``, ``dynamic_batch_size``].\n
            ``one_query`` - sets input shape to [1, max_seq_len]\n
            ``batch`` - sets input shape to [batch_size, max_seq_len]\n
            ``dynamic_batch_size`` - sets batch_size to dynamic range [?, max_seq_len]\n
            Default: ``one_query``.
        :param batch_size: Batch size, required for ``batch`` mode.
            Default: ``None``.
        :param num_candidates_to_score: Number of item ids to calculate scores.
            Could be one of [``None``, ``-1``, ``N``].\n
            ``-1`` - sets candidates_to_score shape to dynamic range [1, ?]\n
            ``N`` - sets candidates_to_score shape to [1, N]\n
            ``None`` - disable candidates_to_score usage\n
            Default: ``None``.
        :param num_threads: Number of CPU threads to use.
            Must be a natural number or ``None``.
            If ``None``, then compiler will set this parameter automatically.
            Default: ``None``.
        :param onnx_path: Save ONNX model to path, if defined.
            Default: ``None``.
        """
        if mode not in get_args(OptimizedModeType):
            msg = f"Parameter ``mode`` could be one of {get_args(OptimizedModeType)}."
            raise ValueError(msg)
        num_candidates_to_score = SasRecCompiled._validate_num_candidates_to_score(num_candidates_to_score)
        if isinstance(model, SasRec):
            lightning_model = model.cpu()
        elif isinstance(model, (str, pathlib.Path)):
            lightning_model = SasRec.load_from_checkpoint(model, map_location=torch.device("cpu"))

        schema = lightning_model._schema
        item_seq_name = schema.item_id_feature_name
        max_seq_len = lightning_model._model.max_len

        batch_size, num_candidates_to_score = SasRecCompiled._get_input_params(
            mode, batch_size, num_candidates_to_score
        )

        item_sequence = torch.zeros((1, max_seq_len)).long()
        padding_mask = torch.zeros((1, max_seq_len)).bool()

        model_input_names = [item_seq_name, "padding_mask"]
        model_dynamic_axes_in_input = {
            item_seq_name: {0: "batch_size", 1: "max_len"},
            "padding_mask": {0: "batch_size", 1: "max_len"},
        }
        if num_candidates_to_score:
            candidates_to_score = torch.zeros((1,)).long()
            model_input_names += ["candidates_to_score"]
            model_dynamic_axes_in_input["candidates_to_score"] = {0: "num_candidates_to_score"}
            model_input_sample = (
                {item_seq_name: item_sequence},
                padding_mask,
                candidates_to_score,
            )
        else:
            model_input_sample = ({item_seq_name: item_sequence}, padding_mask)

        compiled_model = SasRecCompiled._run_model_compilation(
            lightning_model,
            model_input_sample,
            model_input_names,
            model_dynamic_axes_in_input,
            batch_size,
            num_candidates_to_score,
            num_threads,
            onnx_path,
        )

        return cls(compiled_model, schema)



import warnings
from typing import Optional, Union

from replay.utils import PandasDataFrame, PolarsDataFrame, SparkDataFrame

from .base_metric import Metric, MetricsDataFrameLike, MetricsReturnType
from .coverage import Coverage
from .novelty import Novelty
from .recall import Recall
from .surprisal import Surprisal


class OfflineMetrics:
    """
    Designed for efficient calculation of offline metrics provided by the RePlay.
    If you need to calculate multiple metrics for the same input data,
    then using this class is much more efficient than calculating metrics individually.

    For example, you want to calculate several metrics with different parameters.
    When calling offline metrics with the specified metrics,
    the common part of these metrics will be computed only once.

    >>> from replay.metrics import *
    >>> recommendations
       query_id  item_id  rating
    0         1        3    0.6
    1         1        7    0.5
    2         1       10    0.4
    3         1       11    0.3
    4         1        2    0.2
    5         2        5    0.6
    6         2        8    0.5
    7         2       11    0.4
    8         2        1    0.3
    9         2        3    0.2
    10        3        4    1.0
    11        3        9    0.5
    12        3        2    0.1
    >>> groundtruth
       query_id  item_id
    0         1        5
    1         1        6
    2         1        7
    3         1        8
    4         1        9
    5         1       10
    6         2        6
    7         2        7
    8         2        4
    9         2       10
    10        2       11
    11        3        1
    12        3        2
    13        3        3
    14        3        4
    15        3        5
    >>> train
        query_id  item_id
    0         1        5
    1         1        6
    2         1        8
    3         1        9
    4         1        2
    5         2        5
    6         2        8
    7         2       11
    8         2        1
    9         2        3
    10        3        4
    11        3        9
    12        3        2
    >>> base_rec
       query_id  item_id  rating
    0        1        3    0.5
    1        1        7    0.5
    2        1        2    0.7
    3        2        5    0.6
    4        2        8    0.6
    5        2        3    0.3
    6        3        4    1.0
    7        3        9    0.5
    >>> from replay.metrics import Median, ConfidenceInterval, PerUser
    >>> metrics = [
    ...     Precision(2),
    ...     Precision(2, mode=PerUser()),
    ...     Precision(2, mode=Median()),
    ...     Precision(2, mode=ConfidenceInterval(alpha=0.95)),
    ...     Recall(2),
    ...     MAP(2),
    ...     MRR(2),
    ...     NDCG(2),
    ...     HitRate(2),
    ...     RocAuc(2),
    ...     Coverage(2),
    ...     Novelty(2),
    ...     Surprisal(2),
    ... ]
    >>> OfflineMetrics(metrics)(recommendations, groundtruth, train)
    {'Precision@2': 0.3333333333333333,
     'Precision-PerUser@2': {1: 0.5, 2: 0.0, 3: 0.5},
     'Precision-Median@2': 0.5,
     'Precision-ConfidenceInterval@2': 0.32666066409000905,
     'Recall@2': 0.12222222222222223,
     'MAP@2': 0.25,
     'MRR@2': 0.5,
     'NDCG@2': 0.3333333333333333,
     'HitRate@2': 0.6666666666666666,
     'RocAuc@2': 0.3333333333333333,
     'Coverage@2': 0.5555555555555556,
     'Novelty@2': 0.3333333333333333,
     'Surprisal@2': 0.6845351232142715}
    >>> metrics = [
    ...     Precision(2),
    ...     Unexpectedness([1, 2]),
    ...     Unexpectedness([1, 2], mode=PerUser()),
    ... ]
    >>> OfflineMetrics(metrics)(
    ...     recommendations,
    ...     groundtruth,
    ...     train,
    ...     base_recommendations={"ALS": base_rec, "KNN": recommendations}
    ... )
    {'Precision@2': 0.3333333333333333,
     'Unexpectedness_ALS@1': 0.6666666666666666,
     'Unexpectedness_ALS@2': 0.16666666666666666,
     'Unexpectedness_KNN@1': 0.0,
     'Unexpectedness_KNN@2': 0.0,
     'Unexpectedness-PerUser_ALS@1': {1: 1.0, 2: 1.0, 3: 0.0},
     'Unexpectedness-PerUser_ALS@2': {1: 0.5, 2: 0.0, 3: 0.0},
     'Unexpectedness-PerUser_KNN@1': {1: 0.0, 2: 0.0, 3: 0.0},
     'Unexpectedness-PerUser_KNN@2': {1: 0.0, 2: 0.0, 3: 0.0}}
    <BLANKLINE>
    """

    _metrics_call_requirement_map: dict[str, list[str]] = {
        "HitRate": ["ground_truth"],
        "MAP": ["ground_truth"],
        "NDCG": ["ground_truth"],
        "RocAuc": ["ground_truth"],
        "Coverage": ["train"],
        "Novelty": ["train"],
        "Surprisal": ["train"],
        "MRR": ["ground_truth"],
        "Precision": ["ground_truth"],
        "Recall": ["ground_truth"],
    }

    def __init__(
        self,
        metrics: list[Metric],
        query_column: str = "query_id",
        item_column: str = "item_id",
        rating_column: str = "rating",
        category_column: str = "category_id",
        allow_caching: bool = True,
    ):
        """
        :param metrics: (list of metrics): List of metrics to be calculated.
        :param query_column:: (str): The name of the query column.
            Note that you do not need to specify the value of this parameter for each metric separately.
            It is enough to specify the value of this parameter here once.
        :param item_column: (str): The name of the item column.
            Note that you do not need to specify the value of this parameter for each metric separately.
            It is enough to specify the value of this parameter here once.
        :param rating_column: (str): The name of the rating column.
            Note that you do not need to specify the value of this parameter for each metric separately.
            It is enough to specify the value of this parameter here once.
        :param category_column: (str): The name of the category column.
            Note that you do not need to specify the value of this parameter for each metric separately.
            It is enough to specify the value of this parameter here once.

            It is used only for calculating the ``Diversity`` metric.
            If you don't calculate this metric, you can omit this parameter.
        :param allow_caching: (bool): The flag for using caching to optimize calculations.
            Default: ``True``.
        """
        self.unexpectedness_metric: list[Metric] = []
        self.diversity_metric: list[Metric] = []
        self.main_metrics: list[Metric] = []
        self._allow_caching = allow_caching

        for metric in metrics:
            metric.query_column = query_column
            metric.item_column = item_column
            metric.rating_column = rating_column
            if metric.__class__.__name__ in ["Unexpectedness"]:
                self.unexpectedness_metric.append(metric)
            elif metric.__class__.__name__ in ["CategoricalDiversity"]:
                metric.category_column = category_column
                self.diversity_metric.append(metric)
            else:
                self.main_metrics.append(metric)

        self.metrics = self.main_metrics

    def _get_enriched_recommendations(
        self,
        recommendations: Union[SparkDataFrame, PolarsDataFrame],
        ground_truth: Union[SparkDataFrame, PolarsDataFrame],
        train: Optional[Union[SparkDataFrame, PolarsDataFrame]],
    ) -> tuple[dict[str, Union[SparkDataFrame, PolarsDataFrame]], Optional[Union[SparkDataFrame, PolarsDataFrame]]]:
        if len(self.main_metrics) == 0:
            return {}, train
        result_dict = {}
        query_column = self.main_metrics[0].query_column
        item_column = self.main_metrics[0].item_column
        rating_column = self.main_metrics[0].rating_column
        default_metric = Recall(
            topk=2,
            query_column=query_column,
            item_column=item_column,
            rating_column=rating_column,
        )
        is_spark = isinstance(recommendations, SparkDataFrame)
        if is_spark:
            default_metric._check_duplicates_spark(recommendations)
        else:
            default_metric._check_duplicates_polars(recommendations)
        unchanged_recs = recommendations

        result_dict["default"] = default_metric._get_enriched_recommendations(recommendations, ground_truth)

        for metric in self.metrics:
            # find Coverage
            if metric.__class__.__name__ == "Coverage":
                result_dict["Coverage"] = Coverage(
                    topk=2,
                    query_column=query_column,
                    item_column=item_column,
                    rating_column=rating_column,
                )._get_enriched_recommendations(recommendations)

            # find Novelty
            if metric.__class__.__name__ == "Novelty" and train is not None:
                novelty_metric = Novelty(
                    topk=2,
                    query_column=query_column,
                    item_column=item_column,
                    rating_column=rating_column,
                )
                cur_recs = novelty_metric._get_enriched_recommendations(unchanged_recs, train)
                if is_spark:
                    cur_recs = cur_recs.withColumnRenamed("ground_truth", "train")
                else:
                    cur_recs = cur_recs.rename({"ground_truth": "train"})
                cur_recs = metric._rearrange_columns(cur_recs)
                result_dict["Novelty"] = cur_recs

            # find Surprisal
            if metric.__class__.__name__ == "Surprisal" and train is not None:
                result_dict["Surprisal"] = Surprisal(
                    topk=2,
                    query_column=query_column,
                    item_column=item_column,
                    rating_column=rating_column,
                )._get_enriched_recommendations(unchanged_recs, train)

        return result_dict, train

    def _cache_dataframes(self, dataframes: dict[str, SparkDataFrame]) -> None:
        for data in dataframes.values():
            data.cache()

    def _unpersist_dataframes(self, dataframes: dict[str, SparkDataFrame]) -> None:
        for data in dataframes.values():
            data.unpersist()

    def _calculate_metrics(
        self,
        enriched_recs_dict: dict[str, Union[SparkDataFrame, PolarsDataFrame]],
        train: Optional[Union[SparkDataFrame, PolarsDataFrame]] = None,
        is_spark: bool = True,
    ) -> MetricsReturnType:
        result: dict = {}
        for metric in self.metrics:
            metric_args = {}
            if metric.__class__.__name__ == "Coverage" and train is not None:
                metric_args["recs"] = enriched_recs_dict["Coverage"]
                metric_args["train"] = train
            elif metric.__class__.__name__ == "Surprisal":
                metric_args["recs"] = enriched_recs_dict["Surprisal"]
            elif metric.__class__.__name__ == "Novelty":
                metric_args["recs"] = enriched_recs_dict["Novelty"]
            else:
                metric_args["recs"] = enriched_recs_dict["default"]

            if is_spark:
                result.update(metric._spark_compute(**metric_args))
            else:
                result.update(metric._polars_compute(**metric_args))
        return result

    def _check_dataframes_types(
        self,
        recommendations: MetricsDataFrameLike,
        ground_truth: MetricsDataFrameLike,
        train: Optional[MetricsDataFrameLike],
        base_recommendations: Optional[Union[MetricsDataFrameLike, dict[str, MetricsDataFrameLike]]],
    ) -> None:
        types = set()
        types.add(type(recommendations))
        types.add(type(ground_truth))
        if train is not None:
            types.add(type(train))
        if isinstance(base_recommendations, dict):
            for df in base_recommendations.values():
                if not isinstance(df, list):
                    types.add(type(df))
                else:
                    types.add(dict)
                    break
        elif base_recommendations is not None:
            types.add(type(base_recommendations))

        if len(types) != 1:
            msg = "All given data frames must have the same type"
            raise ValueError(msg)

    def _check_query_column_present(
        self,
        dataset: MetricsDataFrameLike,
        query_column: str,
        dataset_name: str,
    ):
        """
        Checks that query column presented in provided dataframe.

        :param dataset: input dataframe.
        :param query_column: name of query column.
        :param dataset_name: name of dataframe.

        :raises KeyError: if query column not found in dataframe.
        """
        if isinstance(dataset, SparkDataFrame):
            dataset_names = dataset.schema.names
        elif isinstance(dataset, (PandasDataFrame, PolarsDataFrame)):
            dataset_names = dataset.columns

        if not isinstance(dataset, dict) and query_column not in dataset_names:
            msg = f"Query column {query_column} is not present in {dataset_name} dataframe"
            raise KeyError(msg)

    def _get_unique_queries(
        self,
        dataset: MetricsDataFrameLike,
        query_column: str,
    ):
        """
        Returns unique queries from provided dataframe.

        :param dataset: input dataframe.
        :param query_column: name of query column.

        :returns: set of unique queries.
        """
        if isinstance(dataset, SparkDataFrame):
            return set(dataset.select(query_column).distinct().toPandas()[query_column])
        elif isinstance(dataset, PandasDataFrame):
            return set(dataset[query_column].unique())
        elif isinstance(dataset, PolarsDataFrame):
            return set(dataset.select(query_column).unique().to_series())
        else:
            return set(dataset.keys())

    def _check_contains(self, queries: set, other_queries: set, dataset_name: str):
        """
        Checks all queries from the first set are presented in the second one.
        Throws warning otherwise.

        :param queries: first set of queries.
        :param other_queries: second set of queries.
        :param dataset_name: name of dataset to specify in warning message.
        """
        if queries.issubset(other_queries) is False:
            warnings.warn(f"{dataset_name} contains queries that are not presented in recommendations")

    def __call__(  # noqa: C901
        self,
        recommendations: MetricsDataFrameLike,
        ground_truth: MetricsDataFrameLike,
        train: Optional[MetricsDataFrameLike] = None,
        base_recommendations: Optional[Union[MetricsDataFrameLike, dict[str, MetricsDataFrameLike]]] = None,
    ) -> dict[str, float]:
        """
        Compute metrics.

        :param recommendations: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict):
            model predictions.
            If DataFrame then it must contains user, item and score columns.
            If dict then key represents user_ids, value represents list of tuple(item_id, score).
        :param ground_truth: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict):
            test data.
            If DataFrame then it must contains user and item columns.
            If dict then key represents user_ids, value represents list of item_ids.
        :param train: (PySpark DataFrame or Polars DataFrame or Pandas DataFrame or dict, optional):
            train data.
            If DataFrame then it must contains user and item columns.
            If dict then key represents user_ids, value represents list of item_ids.
            Default: ``None``.
        :param base_recommendations: (PySpark DataFrame or Polars DataFrame or
            Pandas DataFrame or dict or Dict[str, DataFrameLike]):
            predictions from baseline model.
            If DataFrame then it must contains user, item and score columns.
            If dict then key represents user_ids, value represents list of tuple(item_id, score).
            If ``Unexpectedness`` is not in given metrics list, then you can omit this parameter.
            If it is necessary to calculate the value of metrics on several dataframes,
            then you need to submit a dict(key - name of the data frame, value - DataFrameLike).
            For a better understanding, check out the examples.
            Default: ``None``.

        :return: metric values
        """
        self._check_dataframes_types(recommendations, ground_truth, train, base_recommendations)

        if len(self.main_metrics) > 0:
            query_column = self.main_metrics[0].query_column
        elif len(self.unexpectedness_metric) > 0:
            query_column = self.unexpectedness_metric[0].query_column
        else:
            query_column = self.diversity_metric[0].query_column

        self._check_query_column_present(recommendations, query_column, "recommendations")
        recs_queries = self._get_unique_queries(recommendations, query_column)

        self._check_query_column_present(ground_truth, query_column, "ground_truth")
        self._check_contains(recs_queries, self._get_unique_queries(ground_truth, query_column), "ground_truth")

        if train is not None:
            self._check_query_column_present(train, query_column, "train")
            self._check_contains(recs_queries, self._get_unique_queries(train, query_column), "train")
        if base_recommendations is not None:
            if not isinstance(base_recommendations, dict) or isinstance(
                next(iter(base_recommendations.values())), list
            ):
                base_recommendations = {"base_recommendations": base_recommendations}
            for name, dataset in base_recommendations.items():
                self._check_query_column_present(dataset, query_column, name)
                self._check_contains(recs_queries, self._get_unique_queries(dataset, query_column), name)

        result = {}
        if isinstance(recommendations, (SparkDataFrame, PolarsDataFrame)):
            is_spark = isinstance(recommendations, SparkDataFrame)
            assert isinstance(ground_truth, type(recommendations))
            assert train is None or isinstance(train, type(recommendations))
            enriched_recs_dict, train = self._get_enriched_recommendations(recommendations, ground_truth, train)

            if is_spark and self._allow_caching:
                self._cache_dataframes(enriched_recs_dict)
            result.update(self._calculate_metrics(enriched_recs_dict, train, is_spark))
            if is_spark and self._allow_caching:
                self._unpersist_dataframes(enriched_recs_dict)
        else:  # Calculating metrics in dict format
            current_map: dict[str, Union[PandasDataFrame, dict]] = {
                "ground_truth": ground_truth,
                "train": train,
            }
            for metric in self.metrics:
                args_to_call: dict[str, Union[PandasDataFrame, dict]] = {"recommendations": recommendations}
                for data_name in self._metrics_call_requirement_map[str(metric.__class__.__name__)]:
                    args_to_call[data_name] = current_map[data_name]
                result.update(metric(**args_to_call))
        unexpectedness_result = {}
        diversity_result = {}

        if len(self.unexpectedness_metric) != 0:
            if base_recommendations is None:
                msg = "Can not calculate Unexpectedness because base_recommendations is None"
                raise ValueError(msg)
            first_element = next(iter(base_recommendations.values()))
            if isinstance(base_recommendations, dict) and not isinstance(first_element, list):
                for unexp in self.unexpectedness_metric:
                    for model_name in base_recommendations:
                        cur_result = unexp(recommendations, base_recommendations[model_name])
                        for metric_name in cur_result:
                            splitted = metric_name.split("@")
                            splitted[0] += "_" + model_name
                            unexpectedness_result["@".join(splitted)] = cur_result[metric_name]

        if len(self.diversity_metric) != 0:
            for diversity in self.diversity_metric:
                diversity_result.update(diversity(recommendations))

        return {**result, **unexpectedness_result, **diversity_result}



from typing import NamedTuple, Optional, cast

import torch
from torch.utils.data import Dataset as TorchDataset

from replay.data.nn import (
    MutableTensorMap,
    SequentialDataset,
    TensorMap,
    TorchSequentialDataset,
    TorchSequentialValidationDataset,
)


class SasRecTrainingBatch(NamedTuple):
    """
    Batch of data for training.
    Generated by `SasRecTrainingDataset`.
    """

    query_id: torch.LongTensor
    padding_mask: torch.BoolTensor
    features: TensorMap
    labels: torch.LongTensor
    labels_padding_mask: torch.BoolTensor


class SasRecTrainingDataset(TorchDataset):
    """
    Dataset that generates samples to train SasRec-like model
    """

    def __init__(
        self,
        sequential: SequentialDataset,
        max_sequence_length: int,
        sequence_shift: int = 1,
        sliding_window_step: Optional[None] = None,
        padding_value: Optional[int] = None,
        label_feature_name: Optional[str] = None,
    ) -> None:
        """
        :param sequential: Sequential dataset with training data.
        :param max_sequence_length: Max length of sequence.
        :param sequence_shift: Shift of sequence to predict.
        :param sliding_window_step: A sliding window step.
            If not ``None`` provides iteration over sequences with window.
            Default: ``None``.
        :param padding_value: Value for padding a sequence to match the `max_sequence_length`.
            Default: ``0``.
        :param label_feature_name: Name of label feature in provided dataset.
            If ``None`` set an item_id_feature name from sequential dataset.
            Default: ``None``.
        """
        super().__init__()
        if label_feature_name:
            if label_feature_name not in sequential.schema:
                msg = "Label feature name not found in provided schema"
                raise ValueError(msg)

            if not sequential.schema[label_feature_name].is_cat:
                msg = "Label feature must be categorical"
                raise ValueError(msg)

            if not sequential.schema[label_feature_name].is_seq:
                msg = "Label feature must be sequential"
                raise ValueError(msg)

        self._sequence_shift = sequence_shift
        self._max_sequence_length = max_sequence_length + sequence_shift
        self._label_feature_name = label_feature_name or sequential.schema.item_id_feature_name
        self._schema = sequential.schema

        self._inner = TorchSequentialDataset(
            sequential=sequential,
            max_sequence_length=self._max_sequence_length,
            sliding_window_step=sliding_window_step,
            padding_value=padding_value,
        )

    def __len__(self) -> int:
        return len(self._inner)

    def __getitem__(self, index: int) -> SasRecTrainingBatch:
        query_id, padding_mask, features = self._inner[index]

        assert self._label_feature_name
        labels = features[self._label_feature_name][self._sequence_shift :]
        labels_padding_mask = padding_mask[self._sequence_shift :]

        output_features: MutableTensorMap = {}
        for feature_name in self._schema:
            feature = features[feature_name]
            if self._schema[feature_name].is_seq:
                feature = feature[: -self._sequence_shift]
            output_features[feature_name] = feature

        output_features_padding_mask = padding_mask[: -self._sequence_shift]

        return SasRecTrainingBatch(
            query_id=query_id,
            features=output_features,
            padding_mask=cast(torch.BoolTensor, output_features_padding_mask),
            labels=cast(torch.LongTensor, labels),
            labels_padding_mask=cast(torch.BoolTensor, labels_padding_mask),
        )


class SasRecPredictionBatch(NamedTuple):
    """
    Batch of data for model inference.
    Generated by `SasRecPredictionDataset`.
    """

    query_id: torch.LongTensor
    padding_mask: torch.BoolTensor
    features: TensorMap


class SasRecPredictionDataset(TorchDataset):
    """
    Dataset that generates samples to infer SasRec-like model
    """

    def __init__(
        self,
        sequential: SequentialDataset,
        max_sequence_length: int,
        padding_value: Optional[int] = None,
    ) -> None:
        """
        :param sequential: Sequential dataset with data to make predictions at.
        :param max_sequence_length: Max length of sequence.
        :param padding_value: Value for padding a sequence to match the `max_sequence_length`.
            Default: ``0``.
        """
        self._inner = TorchSequentialDataset(
            sequential=sequential,
            max_sequence_length=max_sequence_length,
            padding_value=padding_value,
        )

    def __len__(self) -> int:
        return len(self._inner)

    def __getitem__(self, index: int) -> SasRecPredictionBatch:
        query_id, padding_mask, features = self._inner[index]
        return SasRecPredictionBatch(
            query_id=query_id,
            padding_mask=padding_mask,
            features=features,
        )


class SasRecValidationBatch(NamedTuple):
    """
    Batch of data for validation.
    Generated by `SasRecValidationDataset`.
    """

    query_id: torch.LongTensor
    padding_mask: torch.BoolTensor
    features: TensorMap
    ground_truth: torch.LongTensor
    train: torch.LongTensor


class SasRecValidationDataset(TorchDataset):
    """
    Dataset that generates samples to infer and validate SasRec-like model
    """

    def __init__(
        self,
        sequential: SequentialDataset,
        ground_truth: SequentialDataset,
        train: SequentialDataset,
        max_sequence_length: int,
        padding_value: Optional[int] = None,
        label_feature_name: Optional[str] = None,
    ):
        """
        :param sequential: Sequential dataset with data to make predictions at.
        :param ground_truth: Sequential dataset with ground truth predictions.
        :param train: Sequential dataset with training data.
        :param max_sequence_length: Max length of sequence.
        :param padding_value: Value for padding a sequence to match the `max_sequence_length`.
            Default: ``0``.
        :param label_feature_name: Name of label feature in provided dataset.
            If ``None`` set an item_id_feature name from sequential dataset.
            Default: ``None``.
        """
        self._inner = TorchSequentialValidationDataset(
            sequential=sequential,
            ground_truth=ground_truth,
            train=train,
            max_sequence_length=max_sequence_length,
            padding_value=padding_value,
            label_feature_name=label_feature_name,
        )

    def __len__(self) -> int:
        return len(self._inner)

    def __getitem__(self, index: int) -> SasRecValidationBatch:
        query_id, padding_mask, features, ground_truth, train = self._inner[index]
        return SasRecValidationBatch(
            query_id=query_id,
            padding_mask=padding_mask,
            features=features,
            ground_truth=ground_truth,
            train=train,
        )
